""" This is the guy that actually modifies the content of the CS
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import zlib
import difflib
import six

from diraccfg import CFG
from DIRAC.Core.Utilities import List, Time
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.Core.Security.ProxyInfo import getProxyInfo

__RCSID__ = "$Id$"


class Modificator(object):
    def __init__(self, rpcClient=False, commiterId="unknown"):
        self.commiterTag = "@@-"
        self.commiterId = commiterId
        self.cfgData = CFG()
        self.rpcClient = None
        if rpcClient:
            self.setRPCClient(rpcClient)

    def loadCredentials(self):
        retVal = getProxyInfo()
        if retVal["OK"]:
            credDict = retVal["Value"]
            self.commiterId = "%s@%s - %s" % (
                credDict["username"],
                credDict["group"],
                Time.dateTime().strftime("%Y-%m-%d %H:%M:%S"),
            )
            return retVal
        return retVal

    def setRPCClient(self, rpcClient):
        self.rpcClient = rpcClient

    def loadFromRemote(self):
        retVal = self.rpcClient.getCompressedData()
        if retVal["OK"]:
            self.cfgData = CFG()
            data = retVal["Value"]
            if six.PY3 and isinstance(data, str):
                data = data.encode(errors="surrogateescape")
            self.cfgData.loadFromBuffer(zlib.decompress(data).decode())
        return retVal

    def getCFG(self):
        return self.cfgData

    def getSections(self, sectionPath):
        return gConfigurationData.getSectionsFromCFG(sectionPath, self.cfgData)

    def getComment(self, sectionPath):
        return gConfigurationData.getCommentFromCFG(sectionPath, self.cfgData)

    def getOptions(self, sectionPath):
        return gConfigurationData.getOptionsFromCFG(sectionPath, self.cfgData)

    def getOptionsDict(self, sectionPath):
        """Gives the options of a CS section in a Python dict with values as
        lists"""

        opts = self.getOptions(sectionPath)
        pathDict = dict((o, self.getValue("%s/%s" % (sectionPath, o))) for o in opts)
        return pathDict

    def getDictRootedAt(self, relpath="", root=""):
        """Gives the configuration rooted at path in a Python dict. The
        result is a Python dictionary that reflects the structure of the
        config file."""

        def getDictRootedAt(path):
            retval = {}
            opts = self.getOptionsDict(path)
            secs = self.getSections(path)
            for k in opts:
                retval[k] = opts[k]
            for i in secs:
                retval[i] = getDictRootedAt(path + "/" + i)
            return retval

        return getDictRootedAt(root + "/" + relpath)

    def getValue(self, optionPath):
        return gConfigurationData.extractOptionFromCFG(optionPath, self.cfgData)

    def sortAlphabetically(self, path, ascending=True):
        cfg = self.__getParentCFG(path, parentLevel=0)
        if cfg:
            if cfg.sortAlphabetically(ascending):
                self.__setCommiter(path)

    def __getParentCFG(self, path, parentLevel=1):
        sectionList = List.fromChar(path, "/")
        cfg = self.cfgData
        try:
            if parentLevel > 0:
                sectionList = sectionList[:-parentLevel]
            for section in sectionList:
                cfg = cfg[section]
            return cfg
        except Exception:
            return False

    def __setCommiter(self, entryPath, cfg=False):
        if not cfg:
            cfg = self.__getParentCFG(entryPath)
        entry = List.fromChar(entryPath, "/")[-1]
        comment = cfg.getComment(entry)
        filteredComment = [line.strip() for line in comment.split("\n") if line.find(self.commiterTag) != 0]
        filteredComment.append("%s%s" % (self.commiterTag, self.commiterId))
        cfg.setComment(entry, "\n".join(filteredComment))

    def setOptionValue(self, optionPath, value):
        levelList = [level.strip() for level in optionPath.split("/") if level.strip() != ""]
        parentPath = "/%s" % "/".join(levelList[:-1])
        optionName = List.fromChar(optionPath, "/")[-1]
        self.createSection(parentPath)
        cfg = self.__getParentCFG(optionPath)
        if not cfg:
            return
        cfg.setOption(optionName, value)
        self.__setCommiter(optionPath, cfg)

    def createSection(self, sectionPath):
        levelList = [level.strip() for level in sectionPath.split("/") if level.strip() != ""]
        currentPath = ""
        cfg = self.cfgData
        createdSection = False
        for section in levelList:
            currentPath += "/%s" % section
            if section not in cfg.listSections():
                cfg.createNewSection(section)
                self.__setCommiter(currentPath)
                createdSection = True
            cfg = cfg[section]
        return createdSection

    def setComment(self, entryPath, value):
        cfg = self.__getParentCFG(entryPath)
        entry = List.fromChar(entryPath, "/")[-1]
        if cfg.setComment(entry, value):
            self.__setCommiter(entryPath)
            return True
        return False

    def existsSection(self, sectionPath):
        sectionList = List.fromChar(sectionPath, "/")
        cfg = self.cfgData
        try:
            for section in sectionList[:-1]:
                cfg = cfg[section]
            return len(sectionList) == 0 or sectionList[-1] in cfg.listSections()
        except Exception:
            return False

    def existsOption(self, optionPath):
        sectionList = List.fromChar(optionPath, "/")
        cfg = self.cfgData
        try:
            for section in sectionList[:-1]:
                cfg = cfg[section]
            return sectionList[-1] in cfg.listOptions()
        except Exception:
            return False

    def renameKey(self, path, newName):
        parentCfg = self.cfgData.getRecursive(path, -1)
        if not parentCfg:
            return False
        pathList = List.fromChar(path, "/")
        oldName = pathList[-1]
        if parentCfg["value"].renameKey(oldName, newName):
            pathList[-1] = newName
            self.__setCommiter("/%s" % "/".join(pathList))
            return True
        else:
            return False

    def copyKey(self, originalKeyPath, newKey):
        parentCfg = self.cfgData.getRecursive(originalKeyPath, -1)
        if not parentCfg:
            return False
        pathList = List.fromChar(originalKeyPath, "/")
        originalKey = pathList[-1]
        if parentCfg["value"].copyKey(originalKey, newKey):
            self.__setCommiter("/%s/%s" % ("/".join(pathList[:-1]), newKey))
            return True
        return False

    def removeOption(self, optionPath):
        if not self.existsOption(optionPath):
            return False
        cfg = self.__getParentCFG(optionPath)
        optionName = List.fromChar(optionPath, "/")[-1]
        return cfg.deleteKey(optionName)

    def removeSection(self, sectionPath):
        if not self.existsSection(sectionPath):
            return False
        cfg = self.__getParentCFG(sectionPath)
        sectionName = List.fromChar(sectionPath, "/")[-1]
        return cfg.deleteKey(sectionName)

    def loadFromBuffer(self, data):
        self.cfgData = CFG()
        self.cfgData.loadFromBuffer(data)

    def loadFromFile(self, filename):
        self.cfgData = CFG()
        self.mergeFromFile(filename)

    def dumpToFile(self, filename):
        with open(filename, "wt") as fd:
            fd.write(str(self.cfgData))

    def mergeFromFile(self, filename):
        cfg = CFG()
        cfg.loadFromFile(filename)
        self.cfgData = self.cfgData.mergeWith(cfg)

    def mergeFromCFG(self, cfg):
        self.cfgData = self.cfgData.mergeWith(cfg)

    def mergeSectionFromCFG(self, sectionPath, cfg):
        parentDict = self.cfgData.getRecursive(sectionPath, -1)
        parentCFG = parentDict["value"]
        secName = [lev.strip() for lev in sectionPath.split("/") if lev.strip()][-1]
        secCFG = parentCFG[secName]
        if not secCFG:
            return False
        mergedCFG = secCFG.mergeWith(cfg)
        parentCFG.deleteKey(secName)
        parentCFG.createNewSection(secName, parentDict["comment"], mergedCFG)
        self.__setCommiter(sectionPath)
        return True

    def __str__(self):
        return str(self.cfgData)

    def commit(self):
        compressedData = zlib.compress(str(self.cfgData).encode(), 9)
        return self.rpcClient.commitNewData(compressedData)

    def getHistory(self, limit=0):
        retVal = self.rpcClient.getCommitHistory(limit)
        if retVal["OK"]:
            return retVal["Value"]
        return []

    def showCurrentDiff(self):
        retVal = self.rpcClient.getCompressedData()
        if retVal["OK"]:
            data = retVal["Value"]
            if six.PY3 and isinstance(data, str):
                data = data.encode(errors="surrogateescape")
            remoteData = zlib.decompress(data).decode().splitlines()
            localData = str(self.cfgData).splitlines()
            return difflib.ndiff(remoteData, localData)
        return []

    def getVersionDiff(self, fromDate, toDate):
        retVal = self.rpcClient.getVersionContents([fromDate, toDate])
        if retVal["OK"]:
            fromData = retVal["Value"][0]
            if six.PY3 and isinstance(fromData, str):
                fromData = fromData.encode(errors="surrogateescape")
            fromData = zlib.decompress(fromData).decode()

            toData = retVal["Value"][1]
            if six.PY3 and isinstance(toData, str):
                toData = toData.encode(errors="surrogateescape")
            toData = zlib.decompress(toData).decode()

            return difflib.ndiff(fromData.split("\n"), toData.split("\n"))
        return []

    def mergeWithServer(self):
        retVal = self.rpcClient.getCompressedData()
        if retVal["OK"]:
            remoteCFG = CFG()
            data = retVal["Value"]
            if six.PY3 and isinstance(data, str):
                data = data.encode(errors="surrogateescape")
            remoteCFG.loadFromBuffer(zlib.decompress(data).decode())
            serverVersion = gConfigurationData.getVersion(remoteCFG)
            self.cfgData = remoteCFG.mergeWith(self.cfgData)
            gConfigurationData.setVersion(serverVersion, self.cfgData)
        return retVal

    def rollbackToVersion(self, version):
        return self.rpcClient.rollbackToVersion(version)

    def updateGConfigurationData(self):
        gConfigurationData.setRemoteCFG(self.cfgData)



""" Class that contains client access to the job monitoring handler. """

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.DEncode import ignoreEncodeWarning
from DIRAC.Core.Utilities.JEncode import strToIntDict
from DIRAC.Core.Utilities.Decorators import deprecated


@createClient("WorkloadManagement/JobMonitoring")
class JobMonitoringClient(Client):
    def __init__(self, **kwargs):

        super(JobMonitoringClient, self).__init__(**kwargs)
        self.setServer("WorkloadManagement/JobMonitoring")

    @ignoreEncodeWarning
    def getJobsStatus(self, jobIDs):
        res = self._getRPC().getJobsStatus(jobIDs)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    def getJobParameters(self, jobIDs, parName=None):
        res = self._getRPC().getJobParameters(jobIDs, parName)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    def getJobsParameters(self, jobIDs, parameters):
        res = self._getRPC().getJobsParameters(jobIDs, parameters)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    def getJobsMinorStatus(self, jobIDs):
        res = self._getRPC().getJobsMinorStatus(jobIDs)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    def getJobsApplicationStatus(self, jobIDs):
        res = self._getRPC().getJobsApplicationStatus(jobIDs)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    def getJobsSites(self, jobIDs):
        res = self._getRPC().getJobsSites(jobIDs)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    @ignoreEncodeWarning
    @deprecated("Use getJobsSummary")
    def getJobsPrimarySummary(self, jobIDs):
        res = self._getRPC().getJobsPrimarySummary(jobIDs)

        # Cast the str keys to int
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res

    def getJobsStates(self, jobIDs):
        res = self._getRPC().getJobsStates(jobIDs)
        if res["OK"]:
            res["Value"] = strToIntDict(res["Value"])
        return res



"""
This Service provides functionality to access and modify the
InstalledComponentsDB database
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six

from DIRAC import S_OK, S_ERROR, gLogger

from DIRAC.FrameworkSystem.DB.InstalledComponentsDB import (
    InstalledComponentsDB,
    Component,
    Host,
    InstalledComponent,
    HostLogging,
)
from DIRAC.Core.DISET.RequestHandler import RequestHandler

__RCSID__ = "$Id$"


class ComponentMonitoringHandler(RequestHandler):
    @classmethod
    def initializeHandler(cls, serviceInfo):
        """
        Handler class initialization
        """

        ComponentMonitoringHandler.doCommit = True

        try:
            ComponentMonitoringHandler.db = InstalledComponentsDB()
        except Exception as e:
            gLogger.error("Could not connect to the database", ": %s" % (e))
            return S_ERROR("Could not connect to the database")

        return S_OK("Initialization went well")

    def __joinInstallationMatch(self, installationFields, componentFields, hostFields):
        matchFields = installationFields
        for key in componentFields:
            matchFields["Component." + key] = componentFields[key]
        for key in hostFields:
            matchFields["Host." + key] = hostFields[key]

        return S_OK(matchFields)

    types_addComponent = [dict]

    def export_addComponent(self, component):
        """
        Creates a new Component object on the database
        component argument should be a dictionary with the Component fields and
        its values
        """

        return ComponentMonitoringHandler.db.addComponent(component)

    types_componentExists = [dict]

    def export_componentExists(self, matchFields):
        """
        Returns whether components matching the given criteria exist
        matchFields argument should be a dictionary with the fields to match
        matchFields accepts fields of the form <Field.bigger> and <Field.smaller>
        to filter using > and < relationships.
        """

        return ComponentMonitoringHandler.db.exists(Component, matchFields)

    types_getComponents = [dict, bool, bool]

    def export_getComponents(self, matchFields, includeInstallations, includeHosts):
        """
        Returns a list of all the Components in the database
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        includeInstallations indicates whether data about the installations in
        which the components takes part is to be retrieved
        includeHosts (only if includeInstallations is set to True) indicates
        whether data about the host in which there are instances of this component
        is to be retrieved
        """

        return ComponentMonitoringHandler.db.getComponents(matchFields, includeInstallations, includeHosts)

    types_updateComponents = [dict, dict]

    def export_updateComponents(self, matchFields, updates):
        """
        Updates Components objects on the database
        matchFields argument should be a dictionary with the fields to match
        (instances matching the fields will be updated) or empty to update all
        the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships updates argument
        updates should be a dictionary with the Component fields and their new
        updated values
        """

        return ComponentMonitoringHandler.db.updateComponents(matchFields, updates)

    types_removeComponents = [dict]

    def export_removeComponents(self, matchFields):
        """
        Removes from the database components that match the given fields
        matchFields argument should be a dictionary with the fields to match or
        empty to remove all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        """

        return ComponentMonitoringHandler.db.removeComponents(matchFields)

    types_addHost = [dict]

    def export_addHost(self, host):
        """
        Creates a new Host object on the database
        host argument should be a dictionary with the Host fields and its values
        """

        return ComponentMonitoringHandler.db.addHost(host)

    types_hostExists = [dict]

    def export_hostExists(self, matchFields):
        """
        Returns whether hosts matching the given criteria exist
        matchFields argument should be a dictionary with the fields to match
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        """

        return ComponentMonitoringHandler.db.exists(Host, matchFields)

    types_getHosts = [dict, bool, bool]

    def export_getHosts(self, matchFields, includeInstallations, includeComponents):
        """
        Returns a list of all the Hosts in the database
        matchFields argument should be a dictionary with the fields to match or
        empty to get all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships
        includeInstallations indicates whether data about the installations in
        which the host takes part is to be retrieved
        includeComponents (only if includeInstallations is set to True) indicates
        whether data about the components installed into this host is to
        be retrieved
        """

        return ComponentMonitoringHandler.db.getHosts(matchFields, includeInstallations, includeComponents)

    types_updateHosts = [dict, dict]

    def export_updateHosts(self, matchFields, updates):
        """
        Updates Hosts objects on the database
        matchFields argument should be a dictionary with the fields to
        match (instances matching the fields will be updated) or empty to update
        all the instances
        matchFields also accepts fields of the form <Field.bigger> and
        <Field.smaller> to filter using > and < relationships updates argument
        should be a dictionary with the Host fields and their new updated values
        updates argument should be a dictionary with the Host fields and
        their new updated values
        """

        return ComponentMonitoringHandler.db.updateHosts(matchFields, updates)

    types_removeHosts = [dict]

    def export_removeHosts(self, matchFields):
        """
        Removes from the database hosts that match the given fields
        matchFields argument should be a dictionary with the fields to match or
        empty to remove all the instances matchFields also accepts fields of the
        form <Field.bigger> and <Field.smaller> to filter
        using > and < relationships
        """

        return ComponentMonitoringHandler.db.removeHosts(matchFields)

    types_addInstallation = [dict, dict, dict, bool]

    def export_addInstallation(self, installation, componentDict, hostDict, forceCreate):
        """
        Creates a new InstalledComponent object on the database
        installation argument should be a dictionary with the InstalledComponent
        fields and its values
        componentDict argument should be a dictionary with the Component fields
        and its values
        hostDict argument should be a dictionary with the Host fields and
        its values
        forceCreate indicates whether a new Component and/or Host should be
        created if the given ones do not exist
        """

        return ComponentMonitoringHandler.db.addInstalledComponent(installation, componentDict, hostDict, forceCreate)

    types_installationExists = [dict, dict, dict]

    def export_installationExists(self, installationFields, componentFields, hostFields):
        """
        Returns whether installations matching the given criteria exist
        installationFields argument should be a dictionary with the fields to
        match for the installation
        componentFields argument should be a dictionary with the fields to match
        for the component installed
        hostFields argument should be a dictionary with the fields to match for
        the host where the installation is made
        """

        matchFields = self.__joinInstallationMatch(installationFields, componentFields, hostFields)["Value"]

        return ComponentMonitoringHandler.db.exists(InstalledComponent, matchFields)

    types_getInstallations = [dict, dict, dict, bool]

    def export_getInstallations(self, installationFields, componentFields, hostFields, installationsInfo):
        """
        Returns a list of installations matching the given criteria
        installationFields argument should be a dictionary with the fields to
        match for the installation
        componentFields argument should be a dictionary with the fields to match
        for the component installed
        hostFields argument should be a dictionary with the fields to match for
        the host where the installation is made
        installationsInfo indicates whether information about the components and
        host taking part in the installation is to be provided
        """

        matchFields = self.__joinInstallationMatch(installationFields, componentFields, hostFields)["Value"]

        return ComponentMonitoringHandler.db.getInstalledComponents(matchFields, installationsInfo)

    types_updateInstallations = [dict, dict, dict, dict]

    def export_updateInstallations(self, installationFields, componentFields, hostFields, updates):
        """
        Updates installations matching the given criteria
        installationFields argument should be a dictionary with the fields to
        match for the installation
        componentFields argument should be a dictionary with the fields to match
        for the component installed or empty to update regardless of component
        hostFields argument should be a dictionary with the fields to match for
        the host where the installation is made or empty to update
        regardless of host
        updates argument should be a dictionary with the Installation fields and
        their new updated values
        """

        matchFields = self.__joinInstallationMatch(installationFields, componentFields, hostFields)["Value"]

        return ComponentMonitoringHandler.db.updateInstalledComponents(matchFields, updates)

    types_removeInstallations = [dict, dict, dict]

    def export_removeInstallations(self, installationFields, componentFields, hostFields):
        """
        Removes installations matching the given criteria
        installationFields argument should be a dictionary with the fields to
        match for the installation
        componentFields argument should be a dictionary with the fields to match
        for the component installed
        hostFields argument should be a dictionary with the fields to match for
        the host where the installation is made
        """

        matchFields = self.__joinInstallationMatch(installationFields, componentFields, hostFields)["Value"]

        return ComponentMonitoringHandler.db.removeInstalledComponents(matchFields)

    types_updateLog = [six.string_types, dict]

    def export_updateLog(self, host, fields):
        """
        Updates the log entry for a given host in the database with the given fields
        host is the name of the machine to which the logging information belongs
        fields is a dictionary where the fields contain the logging information to be stored in the database
        """
        result = ComponentMonitoringHandler.db.exists(HostLogging, {"hostName": host})
        if not result["OK"]:
            return result

        if result["Value"]:
            result = ComponentMonitoringHandler.db.updateLogs({"hostName": host}, fields)
        else:
            fields["hostName"] = host
            result = ComponentMonitoringHandler.db.addLog(fields)

        if not result["OK"]:
            return result

        return S_OK("Logs updated correctly")

    types_getLog = [six.string_types]

    def export_getLog(self, host):
        """
        Retrieves the logging information currently stored for the given host
        """
        result = ComponentMonitoringHandler.db.exists(HostLogging, {"hostName": host})
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Host %s does not exist" % host)

        return ComponentMonitoringHandler.db.getLogs({"hostName": host})

    types_getLogs = []

    def export_getLogs(self):
        """
        Retrieves the logging information currently stored for all hosts
        """
        return ComponentMonitoringHandler.db.getLogs()

    types_removeLogs = [dict]

    def export_removeLogs(self, fields):
        """
        Deletes all the matching logging information
        fields is a dictionary containing the values for the fields
        such that any matching entries in the database should be deleted
        """
        result = ComponentMonitoringHandler.db.exists(HostLogging, fields)
        if not result["OK"]:
            return result
        if not result["Value"]:
            return S_ERROR("Host does not exist")

        return ComponentMonitoringHandler.db.removeLogs(fields)



"""
    This module defines a classs for a generic Workflow Parameter. It also defines
    a ParameterCollection class as a list of parameters as well as an AttributeCollection
    class which is the base class for the main Workflow classes.
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import six
from DIRAC.Core.Workflow.Utility import *

__RCSID__ = "$Id$"

# unbound method, returns indentated string


def indent(indent=0):
    return indent * 2 * " "


class Parameter(object):
    def __init__(
        self,
        name=None,
        value=None,
        type=None,
        linked_module=None,
        linked_parameter=None,
        typein=None,
        typeout=None,
        description=None,
        parameter=None,
    ):
        # the priority to assign values
        # if parameter exists all values taken from there
        # and then owerriten by values taken from the arguments
        if isinstance(parameter, Parameter):
            self.name = parameter.name
            self.type = parameter.type
            self.value = parameter.value
            self.description = parameter.description
            self.linked_module = parameter.linked_module
            self.linked_parameter = parameter.linked_parameter
            self.typein = bool(parameter.typein)
            self.typeout = bool(parameter.typeout)
        else:
            #  default values
            self.name = ""
            self.type = "string"
            self.value = ""
            self.description = ""
            self.linked_module = ""
            self.linked_parameter = ""
            self.typein = False
            self.typeout = False

        if name is not None:
            self.name = name
        if type is not None:
            self.type = type
        if value is not None:
            self.setValue(value)
        if description is not None:
            self.description = description
        if linked_module is not None:
            self.linked_module = linked_module
        if linked_parameter is not None:
            self.linked_parameter = linked_parameter
        if typein is not None:
            self.setInput(typein)
        if typeout is not None:
            self.setOutput(typeout)

    def getName(self):
        return self.name

    def setName(self, n):
        self.name = n  # if collection=None it still will work fine

    def getValue(self):
        return self.value

    def getValueTypeCorrected(self):
        # this method used to generate code for the workflow
        # it NOT used to geterate XML!!!
        if self.isTypeString():
            return '"""' + str(self.value).replace('"', r"\"").replace("'", r"\'") + '"""'
        return self.value

    def setValue(self, value, type_=None):
        if type_ is not None:
            self.setType(type_)
        self.setValueByType(value)

    def setValueByType(self, value):
        type = self.type.lower()  # change the register
        if self.isTypeString():
            self.value = str(value)
        elif type == "float":
            self.value = float(value)
        elif type == "int":
            self.value = int(value)
        elif type == "bool":
            self.value = bool(value)
        else:
            # raise TypeError('Can not assing value '+value+' of unknown type '\
            # + self.type + ' to the Parameter '+ str(self.name))
            # print 'WARNING: we do not have established conversion algorithm to
            # assing value ',value,' of unknown type ',self.type, ' to the Parameter
            # ', str(self.name)
            self.value = value

    def getType(self):
        return self.type

    def setType(self, type_):
        self.type = type_

    def isTypeString(self):
        """returns True if type is the string kind"""
        return self.type.lower() in ("string", "jdl", "option", "parameter", "jdlreqt")

    def getDescription(self):
        return self.description

    def setDescription(self, descr):
        self.description = descr

    def link(self, module, parameter):
        self.linked_module = module
        self.linked_parameter = parameter

    def unlink(self):
        self.linked_module = ""
        self.linked_parameter = ""

    def getLinkedModule(self):
        return self.linked_module

    def getLinkedParameter(self):
        return self.linked_parameter

    def getLink(self):
        # we have 4 possibilities
        # two fields can be filled independently
        # it is possible to fill one field with the valid information
        # spaces shall be ignored ( using strip() function)
        if (self.linked_module is None) or (self.linked_module.strip() == ""):
            if (self.linked_parameter is None) or (self.linked_parameter.strip() == ""):
                # both empty
                return ""
            else:
                # parameter filled
                return self.linked_parameter
        else:
            if (self.linked_parameter is None) or (self.linked_parameter.strip() == ""):
                return self.linked_module
        return self.linked_module + "." + self.linked_parameter

    def isLinked(self):
        if (self.linked_module is None) or (self.linked_module.strip() == ""):
            if (self.linked_parameter is None) or (self.linked_parameter.strip() == ""):
                return False
        return True

    def preExecute(self):
        """method to request watever parameter need to be defined before calling execute method
        returns TRUE if it needs to be done, FALSE otherwise
        PS: parameters with the output status only going to be left out"""
        return (not self.isOutput()) or self.isInput()

    def isInput(self):
        return self.typein

    def isOutput(self):
        return self.typeout

    def setInput(self, i):
        if isinstance(i, six.string_types):
            self.typein = self.__setBooleanFromString(i)
        else:
            self.typein = bool(i)

    def setOutput(self, i):
        if isinstance(i, six.string_types):
            self.typeout = self.__setBooleanFromString(i)
        else:
            self.typeout = bool(i)

    def __setBooleanFromString(self, i):
        return i.upper() == "TRUE"

    def __str__(self):
        return (
            str(type(self))
            + ": name="
            + self.name
            + " value="
            + str(self.getValueTypeCorrected())
            + " type="
            + str(self.type)
            + " linked_module="
            + str(self.linked_module)
            + " linked_parameter="
            + str(self.linked_parameter)
            + " in="
            + str(self.typein)
            + " out="
            + str(self.typeout)
            + " description="
            + str(self.description)
        )

    def toXML(self):
        return (
            '<Parameter name="'
            + self.name
            + '" type="'
            + str(self.type)
            + '" linked_module="'
            + str(self.linked_module)
            + '" linked_parameter="'
            + str(self.linked_parameter)
            + '" in="'
            + str(self.typein)
            + '" out="'
            + str(self.typeout)
            + '" description="'
            + str(self.description)
            + '">'
            + "<value><![CDATA["
            + str(self.getValue())
            + "]]></value>"
            + "</Parameter>\n"
        )

    def compare(self, s):
        if isinstance(s, Parameter):
            return (
                (self.name == s.name)
                and (self.value == s.value)
                and (self.type == s.type)
                and (self.linked_module == s.linked_module)
                and (self.linked_parameter == s.linked_parameter)
                and (self.typein == s.typein)
                and (self.typeout == s.typeout)
                and (self.description == s.description)
            )
        else:
            return False

    def copy(self, parameter):
        if not isinstance(parameter, Parameter):
            raise TypeError("Can not make a copy of object " + str(type(self)) + " from the " + str(type(parameter)))
        self.name = parameter.name
        self.value = parameter.value
        self.type = parameter.type
        self.description = parameter.description
        self.linked_module = parameter.linked_module
        self.linked_parameter = parameter.linked_parameter
        self.typein = parameter.typein
        self.typeout = parameter.typeout

    def createParameterCode(self, ind=0, instance_name=None):
        if (instance_name is None) or (instance_name == ""):
            ret = indent(ind) + self.getName() + " = " + self.getValueTypeCorrected()
        else:
            if self.isLinked():
                ret = indent(ind) + instance_name + "." + self.getName() + " = " + self.getLink()
            else:
                ret = indent(ind) + instance_name + "." + self.getName() + " = " + str(self.getValueTypeCorrected())

        return (
            ret
            + "  # type="
            + self.getType()
            + " in="
            + str(self.isInput())
            + " out="
            + str(self.isOutput())
            + " "
            + self.getDescription()
            + "\n"
        )


class ParameterCollection(list):
    """Parameter collection class representing a list of Parameters"""

    def __init__(self, coll=None):
        list.__init__(self)
        if isinstance(coll, ParameterCollection):
            # makes a deep copy of the parameters
            for v in coll:
                self.append(Parameter(parameter=v))
        elif coll is not None:
            raise TypeError("Can not create object type " + str(type(self)) + " from the " + str(type(coll)))

    def appendOrOverwrite(self, opt):
        index = self.findIndex(opt.getName())
        if index > -1:
            # print "Warning: Overriting Parameter %s = \"%s\" with the value
            # \"%s\""%(self[index].getName(), self[index].getValue(), opt.getValue())
            self[index] = opt
        else:
            list.append(self, opt)

    def append(self, opt):
        if isinstance(opt, ParameterCollection):
            for p in opt:
                self.appendOrOverwrite(p)
        elif isinstance(opt, Parameter):
            self.appendOrOverwrite(opt)
            return opt
        else:
            raise TypeError(
                "Can not append object type "
                + str(type(opt))
                + " to the "
                + str(type(self))
                + ". Parameter type appendable only"
            )

    def appendCopy(self, opt, prefix="", postfix=""):
        if isinstance(opt, ParameterCollection):
            for p in opt:
                self.appendOrOverwrite(Parameter(name=prefix + p.getName() + postfix, parameter=p))
        elif isinstance(opt, Parameter):
            self.appendOrOverwrite(Parameter(name=prefix + opt.getName() + postfix, parameter=opt))
        else:
            raise TypeError(
                "Can not append object type "
                + str(type(opt))
                + " to the "
                + str(type(self))
                + ". Parameter type appendable only"
            )

    def appendCopyLinked(self, opt, prefix="", postfix=""):
        if isinstance(opt, ParameterCollection):
            for p in opt:
                if p.isLinked():
                    self.appendOrOverwrite(Parameter(name=prefix + p.getName() + postfix, parameter=p))
        elif isinstance(opt, Parameter):
            if opt.isLinked():
                self.appendOrOverwrite(Parameter(name=prefix + opt.getName() + postfix, parameter=opt))
        else:
            raise TypeError(
                "Can not append object type "
                + str(type(opt))
                + " to the "
                + str(type(self))
                + ". Parameter type appendable only"
            )

    def setValue(self, name, value, vtype=None):
        """Method finds parameter with the name "name" and if exists its set value
        Returns True if successful
        """
        par = self.find(name)
        if par is None:
            print(
                "ERROR ParameterCollection.setValue() can not find parameter with "
                "the name=%s to set Value=%s" % (name, value)
            )
            return False
        else:
            par.setValue(value, vtype)
            return True

    def getInput(self):
        """Get input linked parameters"""
        return self.get(input=True)

    def getOutput(self):
        """Get output linked parameters"""
        return self.get(output=True)

    def getLinked(self):
        """Get linked parameters"""
        return self.get(input=True, output=True)

    def get(self, input=False, output=False):
        """Get a copy of parameters. If input or output is True, get corresponding
        io type parameters only. Otherwise, get all the parameters
        """
        all = not input and not output

        params = ParameterCollection()
        for p in self:
            OK = False
            if all:
                OK = True
            elif input and p.isInput():
                OK = True
            elif output and p.isOutput():
                OK = True
            if OK:
                params.append(Parameter(parameter=p))

        return params

    def setLink(self, name, module_name, parameter_name):
        """Method finds parameter with the name "name" and if exists its set value
        Returns True if successful
        """
        par = self.find(name)
        if par is None:
            print(
                "ERROR ParameterCollection.setLink() can not find parameter with the name=%s to link it with %s.%s"
                % (name, module_name, parameter_name)
            )
            return False
        else:
            par.link(module_name, parameter_name)
            return True

    def linkUp(self, opt, prefix="", postfix="", objname="self"):
        """This is a GROUP method operates on the 'obj' parameters using only parameters listed in 'opt' list
        Method will link self.parameters with the outer object (self) perameters using prefix and postfix
        for example if we want to link module instance with the step or step instance with the workflow
        opt - ParameterCollection or single Parameter (WARNING: used as reference to get a names, opt is not changing!!!)
        opt ALSO can be a list of string with the names of parameters to link
        objname - name of the object to connect with, usually 'self'
        """
        if isinstance(opt, ParameterCollection):
            # if parameter in the list opt is not present in the self
            # we are going to ignore this
            for p in opt:
                par = self.find(p.getName())
                if par is None:
                    print(
                        "WARNING ParameterCollection.linkUp can not find parameter with the name=",
                        p.getName(),
                        " IGNORING",
                    )
                else:
                    par.link(objname, prefix + p.getName() + postfix)
        elif isinstance(opt, Parameter):
            self.setLink(opt.getName(), objname, prefix + opt.getName() + postfix)
        elif isinstance(opt, list) and isinstance(opt[0], str):
            for s in opt:
                par = self.find(s)
                if par is None:
                    print("ERROR ParameterCollection.linkUp() can not find parameter with the name=%s" % (s))
                else:
                    par.link(objname, prefix + p.getName() + postfix)
        elif isinstance(opt, str):
            par = self.find(opt)
            if par is None:
                print("ERROR ParameterCollection.linkUp() can not find parameter with the name=%s" % (par))
            else:
                par.link(objname, prefix + par.getName() + postfix)
        else:
            raise TypeError("Can not link object type " + str(type(opt)) + " to the " + str(type(self)) + ".")

    def unlink(self, opt):
        """This is a GROUP method operates on the 'obj' parameters using only parameters listed in 'opt' list
        Method will unlink some self.parameters
        opt - ParameterCollection or single Parameter (WARNING: used as reference to get a names, opt is not changing!!!)
        opt ALSO can be a list of string with the names of parameters to link
        objname - name of the object to connect with, usually 'self'
        """
        if isinstance(opt, ParameterCollection):
            # if parameter in the list opt is not present in the self
            # we are going to ignore this
            for p in opt:
                par = self.find(p.getName())
                if par is None:
                    print(
                        "WARNING ParameterCollection.linkUp can not find parameter with the name=",
                        p.getName(),
                        " IGNORING",
                    )
                else:
                    par.unlink()
        elif isinstance(opt, Parameter):
            opt.unlink()
        elif isinstance(opt, list) and isinstance(opt[0], str):
            for s in opt:
                par = self.find(s)
                if par is None:
                    print("ERROR ParameterCollection.unlink() can not find parameter with the name=%s" % (s))
                else:
                    par.unlink()
        elif isinstance(opt, str):
            par = self.find(opt)
            if par is None:
                print("ERROR ParameterCollection.unlink() can not find parameter with the name=%s" % (s))
            else:
                par.unlink()
        else:
            raise TypeError("Can not unlink object type " + str(type(opt)) + " to the " + str(type(self)) + ".")

    def removeAllParameters(self):
        self[:] = []

    def remove(self, name_or_ind):
        """Removes a parameter given its name, or the index (the latter is not suggested), and only if it exists

        If there are 2 parameters with the same name, only the first will be removed
        """
        if isinstance(name_or_ind, list) and isinstance(name_or_ind[0], str):
            for s in name_or_ind:
                par = self.find(s)
                if par is None:
                    print("ERROR ParameterCollection.remove() can not find parameter with the name=%s" % (s))
                else:
                    index = self.findIndex(s)
                    if index > -1:
                        del self[index]

        elif isinstance(name_or_ind, str):  # we give a name
            index = self.findIndex(name_or_ind)
        elif isinstance(name_or_ind, int):  # we give the index
            index = name_or_ind

        if index > -1:
            del self[index]

    def find(self, name_or_ind):
        """Method to find Parameters
        Return: Parameter"""
        # work for index as well as for the string
        if isinstance(name_or_ind, str):  # we given name
            for v in self:
                if v.getName() == name_or_ind:
                    return v
            return None

        elif isinstance(name_or_ind, six.integer_types):  # we given index
            return self[name_or_ind]
        return self[int(name_or_ind)]

    def findLinked(self, name_or_ind, linked_status=True):
        """Method to find Parameters
        if linked_status is True it returns only linked Var from the list
        if linked_status is False it returns only NOTlinked Var from the list
        Return: Parameter"""
        v = self.find(name_or_ind)
        if (v is not None) and (v.isLinked() != linked_status):
            return None
        return v

    def findIndex(self, name):
        for i, v in enumerate(self):
            if v.getName() == name:
                return i
        return -1

    def getParametersNames(self):
        return [v.getName() for v in self]

    def compare(self, s):
        # we comparing parameters only, the attributes will be compared in hierarchy above
        # we ignore the position of the Parameter in the list
        # we assume that names of the Parameters are DIFFERENT otherwise we have to change algorithm!!!
        if (not isinstance(s, ParameterCollection)) or (len(s) != len(self)):
            return False
        for v in self:
            for i in s:
                if v.getName() == i.getName() and v.compare(i):
                    break
                else:
                    return False
            return True

    def __str__(self):
        ret = [str(type(self)) + ":"]
        for v in self:
            ret += [str(v)]
        return "\n".join(ret + [""])

    def toXML(self):
        return "".join(v.toXML() for v in self)

    def createParametersCode(self, indent=0, instance_name=None):
        return "".join(v.createParameterCode(indent, instance_name) for v in self if v.preExecute())

    def resolveGlobalVars(self, wf_parameters=None, step_parameters=None):
        """This function resolves global parameters of type @{value} within the ParameterCollection"""
        recurrency_max = 12
        for v in self:
            recurrency = 0
            skip_list = []
            substitute_vars = getSubstitute(v.value)
            while True:
                for substitute_var in substitute_vars:

                    # looking in the current scope
                    v_other = self.find(substitute_var)

                    # looking in the scope of step instance
                    if v_other is None and step_parameters is not None:
                        v_other = step_parameters.findLinked(substitute_var, False)

                    # looking in the scope of workflow
                    if v_other is None and wf_parameters is not None:
                        v_other = wf_parameters.findLinked(substitute_var, False)

                    # finally the action itself
                    if v_other is not None and not v_other.isLinked():
                        v.value = substitute(v.value, substitute_var, v_other.value)
                    elif v_other is not None:
                        print("Leaving %s variable for dynamic resolution" % substitute_var)
                        skip_list.append(substitute_var)
                    else:  # if nothing helped tough!
                        print("Can not resolve ", substitute_var, str(v))

                recurrency += 1
                if recurrency > recurrency_max:
                    # must be an exception
                    print("ERROR! reached maximum recurrency level", recurrency, "within the parameter ", str(v))
                    if step_parameters is None:
                        if wf_parameters is None:
                            print("on the level of Workflow")
                        else:
                            print("on the level of Step")
                    else:
                        if wf_parameters is not None:
                            print("on the level of Module")
                    break
                else:
                    substitute_vars = getSubstitute(v.value, skip_list)
                    if not substitute_vars:
                        break


class AttributeCollection(dict):
    """Attribute Collection class contains Parameter Collection as a data member"""

    def __init__(self):
        dict.__init__(self)
        self.parameters = None
        self.parent = None

    def __str__(self):
        ret = ""
        for v in self.keys():
            ret = ret + v + " = " + str(self[v]) + "\n"
        return ret

    def toXMLString(self):
        return self.toXML()

    def toXMLFile(self, filename):
        with open(filename, "w+") as f:
            sarray = self.toXML()
            for element in sarray:
                f.write(element)
        return

    def toXML(self):
        ret = ""
        for v in sorted(self):
            if v == "parent":
                continue  # doing nothing
            elif v in ["body", "description"]:
                ret = ret + "<" + v + "><![CDATA[" + str(self[v]) + "]]></" + v + ">\n"
            else:
                ret = ret + "<" + v + ">" + str(self[v]) + "</" + v + ">\n"
        return ret

    def addParameter(self, opt, prefix="", postfix=""):
        self.parameters.appendCopy(opt, prefix, postfix)

    def addParameterLinked(self, opt, prefix="", postfix=""):
        self.parameters.appendCopyLinked(opt, prefix, postfix)

    def linkUp(self, opt, prefix="", postfix="", objname="self"):
        self.parameters.linkUp(opt, prefix, postfix, objname)

    def unlink(self, opt):
        self.parameters.unlink(opt)

    def removeParameter(self, name_or_ind):
        self.parameters.remove(name_or_ind)

    def removeAllParameters(self):
        self.parameters.removeAllParameters()

    def findParameter(self, name_or_ind):
        return self.parameters.find(name_or_ind)

    def findParameterIndex(self, ind):
        return self.parameters.findIndex(ind)

    def compareParameters(self, s):
        return self.parameters.compare(s)

    def setValue(self, name, value, type_=None):
        if not self.parameters.setValue(name, value, type_):
            print(" in the object=", type(self), "with name=", self.getName(), "of type=", self.getType())

    def setLink(self, name, module_name, parameter_name):
        if not self.parameters.setLink(name, module_name, parameter_name):
            print(" in the object=", type(self), "with name=", self.getName(), "of type=", self.getType())

    def compare(self, s):
        return (self == s) and self.parameters.compare(s.parameters)

    def setParent(self, parent):
        self.parent = parent

    def getParent(self):
        return self.parent

    # ------------- common functions -----------
    def setName(self, name):
        self["name"] = name

    def getName(self):
        return self.get("name", "")

    def setType(self, att_type):
        self["type"] = att_type

    def getType(self):
        return self.get("type", "")

    def setRequired(self, required):
        self["required"] = required

    def getRequired(self):
        return self["required"]

    def setDescription(self, description):
        self["description"] = description

    def getDescription(self):
        return self["description"]

    def setDescrShort(self, descr_short):
        self["descr_short"] = descr_short

    def getDescrShort(self):
        return self["descr_short"]

    def setBody(self, body):
        self["body"] = body

    def getBody(self):
        return self["body"]

    def setOrigin(self, origin):
        self["origin"] = origin

    def getOrigin(self):
        return self["origin"]

    def setVersion(self, ver):
        self["version"] = ver

    def getVersion(self):
        return self["version"]

    def resolveGlobalVars(self, wf_parameters=None, step_parameters=None):
        self.parameters.resolveGlobalVars(wf_parameters, step_parameters)



#! /usr/bin/env python
"""
Change status of replica of a given file or a list of files at a given Storage Element
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC import exit as DIRACExit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(("LocalFile: Path to local file containing LFNs", "LFN:       Logical File Name"))
    Script.registerArgument(" SE:        Storage Element")
    Script.registerArgument(" status:    status")
    Script.parseCommandLine()

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    catalog = FileCatalog()
    import os

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    inputFileName, se, newStatus = Script.getPositionalArgs(group=True)

    if os.path.exists(inputFileName):
        inputFile = open(inputFileName, "r")
        string = inputFile.read()
        lfns = string.splitlines()
        inputFile.close()
    else:
        lfns = [inputFileName]

    res = catalog.getReplicas(lfns, True)
    if not res["OK"]:
        print(res["Message"])
        DIRACExit(-1)
    replicas = res["Value"]["Successful"]

    lfnDict = {}
    for lfn in lfns:
        lfnDict[lfn] = {}
        lfnDict[lfn]["SE"] = se
        lfnDict[lfn]["Status"] = newStatus
        lfnDict[lfn]["PFN"] = replicas[lfn][se]

    res = catalog.setReplicaStatus(lfnDict)
    if not res["OK"]:
        print("ERROR:", res["Message"])
    if res["Value"]["Failed"]:
        print("Failed to update %d replica status" % len(res["Value"]["Failed"]))
    if res["Value"]["Successful"]:
        print("Successfully updated %d replica status" % len(res["Value"]["Successful"]))


if __name__ == "__main__":
    main()



""" X509Certificate is a class for managing X509 certificates

Proxy RFC: https://tools.ietf.org/html/rfc38200

X509RFC: https://tools.ietf.org/html/rfc5280

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import os
import random
import time

import M2Crypto

from io import open

from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.Core.Security.m2crypto import asn1_utils
from DIRAC.Core.Utilities.Decorators import executeOnlyIf

# Decorator to execute the method only of the certificate has been loaded
executeOnlyIfCertLoaded = executeOnlyIf("_certLoaded", S_ERROR(DErrno.ENOCERT))


class X509Certificate(object):
    """The X509Certificate object represents ... a X509Certificate.

    It is a wrapper around a lower level implementation (M2Crypto in this case) of a certificate.
    In theory, tt can be a host or user certificate. Also, a proxy certificate is a X509Certificate,
    however it is useless without all the chain of issuers.
    That's why one has the X509Chain.

    In practice, X509Certificate is just used for checking  if the host certificate has expired.
    This class will most probably disappear once we get ride of pyGSI. After all, a X509Certificate
    is nothing but a X509Chain of length 1.

    Note that the SSL connection itself does not use this class, it gives directly the certificate to the library
    """

    def __init__(self, x509Obj=None, certString=None):
        """
        Constructor.
        You can give either nothing, or the x509Obj or the certString

        :param x509Obj: (optional) certificate instance
        :type x509Obj: M2Crypto.X509.X509
        :param certString: text representation of certificate
        :type certString: String

        """

        self._certLoaded = False
        if x509Obj:
            self.__certObj = x509Obj
            self._certLoaded = True
        elif certString:
            self.loadFromString(certString)

    # Pylint is surprisingly picky here, so remove that warning
    # pylint: disable=protected-access
    @classmethod
    def generateProxyCertFromIssuer(cls, x509Issuer, x509ExtensionStack, proxyKey, lifetime=3600):
        """This class method is meant to generate a new X509Certificate out of an existing one.
        Basically, it generates a proxy... However, you can't have a proxy certificate working on
        its own, you need all the chain of certificates. This method is meant to be called
        only by the X509Chain class.

        Inspired from https://github.com/eventbrite/m2crypto/blob/master/demo/x509/ca.py#L45

        :param x509Issuer: X509Certificate instance from which we generate the next one
        :param x509ExtensionStack: M2Crypto.X509.X509_Extension_Stack object to add to the new certificate.
                                   It contains all the X509 extensions needed for the proxy (e.g. DIRAC group).
                                   See ~X509Chain.__getProxyExtensionList
        :param proxyKey: a M2Crypto.EVP.PKey instance with private and public key
        :param lifetime: duration of the proxy in second. Default 3600


        :returns: a new X509Certificate

        """
        proxyCert = cls()

        proxyCert.__certObj = M2Crypto.X509.X509()

        # According to the proxy RFC, the serial
        # number just need to be uniqu among the proxy generated by the issuer.
        # The random module of python will be good enough for that
        serial = int(random.random() * 10 ** 10)
        proxyCert.__certObj.set_serial_number(serial)

        # No easy way to deep-copy certificate subject, since they are swig object

        # We basically get a string like 'O=Dirac Computing, O=CERN, CN=MrUser'
        # So we split it, and then re-add each entry after the other.

        proxySubject = M2Crypto.X509.X509_Name()

        issuerSubjectObj = x509Issuer.__certObj.get_subject()
        issuerSubjectParts = issuerSubjectObj.as_text().split(", ")

        for isPart in issuerSubjectParts:
            nid, val = isPart.split("=", 1)
            proxySubject.add_entry_by_txt(field=nid, type=M2Crypto.ASN1.MBSTRING_ASC, entry=val, len=-1, loc=-1, set=0)

        # Finally we add a random Common Name  component. And we might as well use the serial.. :)
        proxySubject.add_entry_by_txt(
            field="CN", type=M2Crypto.ASN1.MBSTRING_ASC, entry=str(serial), len=-1, loc=-1, set=0
        )

        proxyCert.__certObj.set_subject(proxySubject)

        # We now add all the extensions we wish to add
        for extension in x509ExtensionStack:
            proxyCert.__certObj.add_ext(extension)

        proxyCert.__certObj.set_issuer(issuerSubjectObj)

        # According to the X509 RFC, we are safe if we just copy the version
        # number from the issuer certificate
        proxyCert.__certObj.set_version(x509Issuer.__certObj.get_version())

        proxyCert.__certObj.set_pubkey(proxyKey)

        # Set the start of the validity a bit in the past
        # to be sure to be able to use it right now
        proxyNotBefore = M2Crypto.ASN1.ASN1_UTCTIME()
        proxyNotBefore.set_time(int(time.time()) - 900)
        proxyCert.__certObj.set_not_before(proxyNotBefore)

        # Set the end date of the validity according to the lifetime
        proxyNotAfter = M2Crypto.ASN1.ASN1_UTCTIME()
        proxyNotAfter.set_time(int(time.time()) + lifetime)
        proxyCert.__certObj.set_not_after(proxyNotAfter)

        # Finally set it as loaded. Care that it is not yet signed !!
        proxyCert._certLoaded = True

        return S_OK(proxyCert)

    def load(self, certificate):
        """Load an x509 certificate either from a file or from a string

        :param certificate: path to the file or PEM encoded string

        :returns: S_OK on success, otherwise S_ERROR
        """

        if os.path.exists(certificate):
            return self.loadFromFile(certificate)

        return self.loadFromString(certificate)

    def loadFromFile(self, certLocation):
        """
         Load a x509 cert from a pem file

         :param certLocation: path to the certificate file

        :returns: S_OK / S_ERROR.

        """
        try:
            with open(certLocation, "rb") as fd:
                pemData = fd.read()
                return self.loadFromString(pemData)
        except IOError:
            return S_ERROR(DErrno.EOF, "Can't open %s file" % certLocation)

    def loadFromString(self, pemData):
        """
        Load a x509 cert from a string containing the pem data

        :param pemData: pem encoded string

        :returns: S_OK / S_ERROR
        """
        try:
            self.__certObj = M2Crypto.X509.load_cert_string(pemData, M2Crypto.X509.FORMAT_PEM)
        except Exception as e:
            return S_ERROR(DErrno.ECERTREAD, "Can't load pem data: %s" % e)

        self._certLoaded = True
        return S_OK()

    @executeOnlyIfCertLoaded
    def hasExpired(self):
        """
        Check if the loaded certificate is still valid

        :returns: S_OK( True/False )/S_ERROR
        """

        res = self.getNotAfterDate()
        if not res["OK"]:
            return res

        notAfter = res["Value"]
        now = datetime.datetime.utcnow()

        return S_OK(notAfter < now)

    @executeOnlyIfCertLoaded
    def getNotAfterDate(self):
        """
        Get not after date of a certificate

        :returns: S_OK( datetime )/S_ERROR
        """

        notAfter = self.__certObj.get_not_after().get_datetime()

        # M2Crypto does things correctly by setting a timezone info in the datetime
        # However, we do not in DIRAC, and so we can't compare the dates.
        # We have to remove the timezone info from M2Crypto
        notAfter = notAfter.replace(tzinfo=None)

        return S_OK(notAfter)

    @executeOnlyIfCertLoaded
    def getNotBeforeDate(self):
        """
        Get not before date of a certificate

        :returns: S_OK( datetime )/S_ERROR

        """
        return S_OK(self.__certObj.get_not_before().get_datetime())

    # @executeOnlyIfCertLoaded
    # def setNotBefore(self, notbefore):
    #   """
    #     Set not before date of a certificate This method is not meant to be used, but to generate a proxy.

    #     :returns: S_OK/S_ERROR
    #   """
    #   self.__certObj.set_not_before(notbefore)
    #   return S_OK()

    @executeOnlyIfCertLoaded
    def getSubjectDN(self):
        """
        Get subject DN

        :returns: S_OK( string )/S_ERROR
        """
        return S_OK(str(self.__certObj.get_subject()))

    @executeOnlyIfCertLoaded
    def getIssuerDN(self):
        """
        Get issuer DN

        :returns: S_OK( string )/S_ERROR
        """
        return S_OK(str(self.__certObj.get_issuer()))

    @executeOnlyIfCertLoaded
    def getSubjectNameObject(self):
        """
        Get subject name object

        :returns: S_OK( X509Name )/S_ERROR
        """
        return S_OK(self.__certObj.get_subject())

    # The following method is in pyGSI,
    # but are only used by the pyGSI SSL implementation
    # So I do not really need them

    # @executeOnlyIfCertLoaded
    # def getIssuerNameObject(self):
    #   """
    #     Get issuer name object

    #     :returns: S_OK( X509Name )/S_ERROR
    #   """
    #   return S_OK(self.__certObj.get_issuer())

    @executeOnlyIfCertLoaded
    def getPublicKey(self):
        """
        Get the public key of the certificate

        :returns: S_OK(M2crypto.EVP.PKey)

        """
        return S_OK(self.__certObj.get_pubkey())

    @executeOnlyIfCertLoaded
    def getSerialNumber(self):
        """
        Get certificate serial number

        :returns: S_OK( serial )/S_ERROR
        """
        return S_OK(self.__certObj.get_serial_number())

    @executeOnlyIfCertLoaded
    def sign(self, key, algo):
        """
        Sign the cerificate using provided key and algorithm.

        :param key: M2crypto.EVP.PKey object with private and public key
        :param algo: algorithm to sign the certificate

        :returns: S_OK/S_ERROR
        """
        try:
            self.__certObj.sign(key, algo)
        except Exception as e:
            return S_ERROR(repr(e))

        return S_OK()

    @executeOnlyIfCertLoaded
    def getDIRACGroup(self, ignoreDefault=False):
        """
        Get the dirac group if present

        If no group is found in the certificate, we query the CS to get the default group
        for the given user. This can be disabled using the ignoreDefault parameter

        Note that the lookup in the CS only can work for a proxy of first generation,
        since we search based on the issuer DN

        :param ignoreDefault: if True, do not lookup the CS

        :returns: S_OK(group name/bool)
        """
        try:
            return S_OK(asn1_utils.decodeDIRACGroup(self.__certObj))
        except LookupError:
            pass

        if ignoreDefault:
            return S_OK(False)

        # And here is the flaw :)
        result = self.getIssuerDN()
        if not result["OK"]:
            return result
        return Registry.findDefaultGroupForDN(result["Value"])

    @executeOnlyIfCertLoaded
    def hasVOMSExtensions(self):
        """
        Has voms extensions

        :returns: S_OK(bool) if voms extensions are found
        """

        # `get_ext` would be the correct thing to do.
        # However, it does not work for the moment, as the extension
        # is not registered with an alias
        # https://gitlab.com/m2crypto/m2crypto/issues/231
        # try:
        #   self.__certObj.get_ext('vomsExtensions')
        #   return S_OK(True)
        # except LookupError:
        #   # no extension found
        #   pass

        return S_OK(asn1_utils.hasVOMSExtension(self.__certObj))

    @executeOnlyIfCertLoaded
    def getVOMSData(self):
        """
        Get voms extensions data

        :returns: S_ERROR/S_OK(dict). For the content of the dict,
              see :py:func:`~DIRAC.Core.Security.m2crypto.asn1_utils.decodeVOMSExtension`
        """
        try:
            vomsExt = asn1_utils.decodeVOMSExtension(self.__certObj)
            return S_OK(vomsExt)
        except LookupError:
            return S_ERROR(DErrno.EVOMS, "No VOMS data available")

    @executeOnlyIfCertLoaded
    def generateProxyRequest(self, bitStrength=1024, limited=False):
        """
        Generate a proxy request. See :py:class:`DIRAC.Core.Security.m2crypto.X509Request.X509Request`

        In principle, there is no reason to have this here, since a the X509Request is independant of
        the  509Certificate  when generating it. The only reason is to check whether the current Certificate
        is limited or not.

        :param bitStrength: strength of the key
        :param limited: if True or if the current certificate is limited (see proxy RFC),
                        creates a request for a limited proxy

        :returns: S_OK( :py:class:`DIRAC.Core.Security.m2crypto.X509Request.X509Request` ) / S_ERROR
        """
        if not limited:
            # We check whether "limited proxy" is in the subject
            subj = self.__certObj.get_subject()
            # M2Crypto does not understand the [-1] syntax...
            lastEntry = subj[len(subj) - 1]
            if lastEntry.get_data() == "limited proxy":
                limited = True

        # The import is done here to avoid circular import
        # X509Certificate -> X509Request -> X509Chain -> X509Certificate
        from DIRAC.Core.Security.m2crypto.X509Request import X509Request

        req = X509Request()
        req.generateProxyRequest(bitStrength=bitStrength, limited=limited)

        return S_OK(req)

    @executeOnlyIfCertLoaded
    def getRemainingSecs(self):
        """
        Get remaining lifetime in secs

        :returns: S_OK(remaining seconds)
        """
        notAfter = self.getNotAfterDate()["Value"]
        now = datetime.datetime.utcnow()
        remainingSeconds = max(0, int((notAfter - now).total_seconds()))

        return S_OK(remainingSeconds)

    @executeOnlyIfCertLoaded
    def getExtensions(self):
        """
        Get a decoded list of extensions

        :returns: S_OK( list of tuple (extensionName, extensionValue))
        """
        extList = []
        for i in range(self.__certObj.get_ext_count()):
            sn = self.__certObj.get_ext_at(i).get_name()
            try:
                value = self.__certObj.get_ext_at(i).get_value()
            except Exception:
                value = "Cannot decode value"
            extList.append((sn, value))

        return S_OK(sorted(extList))

    @executeOnlyIfCertLoaded
    def verify(self, pkey):
        """
        Verify the signature of the certificate using the public key provided

        :param pkey: ~M2Crypto.EVP.PKey object

        :returns: S_OK(bool) where the boolean shows the success of the verification
        """
        ret = self.__certObj.verify(pkey)
        return S_OK(ret == 1)

    @executeOnlyIfCertLoaded
    def asPem(self):
        """
        Return certificate as PEM string

        :returns: pem string
        """
        return self.__certObj.as_pem()

    @executeOnlyIfCertLoaded
    def getExtension(self, name):
        """
        Return X509 Extension with given name

        :param name: name of the extension

        :returns: S_OK with M2Crypto.X509.X509_Extension object, or S_ERROR
        """
        try:
            ext = self.__certObj.get_ext(name)
        except LookupError as e:
            return S_ERROR(e)
        return S_OK(ext)



""" KeystoneClient class encapsulates the work with the keystone service interface
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import requests

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Time import fromString, dateTime

__RCSID__ = "$Id$"


class KeystoneClient(object):
    def __init__(self, url, parameters):
        self.log = gLogger.getSubLogger("Keystone")
        self.url = url
        self.apiVersion = None
        if "v3" in url:
            self.apiVersion = 3
        if "v2" in url:
            self.apiVersion = 2
        if self.apiVersion is None:
            # Assume v2.0
            self.apiVersion = 2
            self.url = self.url.rstrip("/") + "/v2.0"
        self.parameters = parameters
        self.token = None
        self.expires = None
        self.project = self.parameters.get("Tenant", self.parameters.get("Project"))
        self.projectID = self.parameters.get("ProjectID")
        self.computeURL = None
        self.imageURL = None
        self.networkURL = None
        self.caPath = self.parameters.get("CAPath", True)

        self.valid = False
        result = self.initialize()
        if result["OK"]:
            self.valid = True
        else:
            gLogger.error("Keystone initialization failed: %s" % result["Message"])

    def initialize(self):
        """Initialize the Keystone object obtaining the corresponding token

        :return: S_OK/S_ERROR
        """

        self.log.debug("Initializing for API version %d" % self.apiVersion)

        result = self.getToken()
        if not result["OK"]:
            return result

        # If the tenant is not specified, try to get it and obtain the tenant specific token
        if not self.project:
            result = self.getTenants()
            if not result["OK"]:
                return result
            if result["Value"]:
                self.project, self.projectID = result["Value"][0]
                result = self.getToken(force=True)
                if not result["OK"]:
                    return result

        return S_OK()

    def getToken(self, force=False):
        """Get the Keystone token

        :param force: flag to force getting the token if even there is one in the cache
        :return: S_OK(token) or S_ERROR
        """

        if self.token is not None and not force:
            if self.expires and (self.expires - dateTime()).seconds > 300:
                return S_OK(self.token)

        if self.apiVersion == 2:
            result = self.__getToken2()
        else:
            result = self.__getToken3()
        return result

    def __getToken2(self):
        """Get the Keystone token for the version v2 of the keystone service

        :return: S_OK(token) or S_ERROR
        """

        user = self.parameters.get("User")
        password = self.parameters.get("Password")
        authArgs = {}
        if user and password:
            authDict = {"auth": {"passwordCredentials": {"username": user, "password": password}}}
            if self.project:
                authDict["auth"]["tenantName"] = self.project
        elif self.parameters.get("Auth") == "voms":
            authDict = {"auth": {"voms": True}}
            if self.project:
                authDict["auth"]["tenantName"] = self.project

            if self.parameters.get("Proxy"):
                authArgs["cert"] = self.parameters.get("Proxy")

        try:
            result = requests.post(
                "%s/tokens" % self.url,
                headers={"Content-Type": "application/json"},
                json=authDict,
                verify=self.caPath,
                **authArgs
            )
        except Exception as exc:
            return S_ERROR("Exception getting keystone token: %s" % str(exc))

        output = result.json()

        if result.status_code in [400, 401]:
            message = "None"
            if "error" in output:
                message = output["error"].get("message")
            return S_ERROR("Authorization error: %s" % message)

        self.token = str(output["access"]["token"]["id"])
        expires = fromString(str(output["access"]["token"]["expires"]).replace("T", " ").replace("Z", ""))
        issued = fromString(str(output["access"]["token"]["issued_at"]).replace("T", " ").replace("Z", ""))
        self.expires = dateTime() + (expires - issued)

        self.projectID = output["access"]["token"]["tenant"]["id"]

        for endpoint in output["access"]["serviceCatalog"]:
            if endpoint["type"] == "compute":
                self.computeURL = str(endpoint["endpoints"][0]["publicURL"])
            elif endpoint["type"] == "image":
                self.imageURL = str(endpoint["endpoints"][0]["publicURL"])
            elif endpoint["type"] == "network":
                self.networkURL = str(endpoint["endpoints"][0]["publicURL"])
        return S_OK(self.token)

    def __getToken3(self):
        """Get the Keystone token for the version v3 of the keystone service

        :return: S_OK(token) or S_ERROR
        """

        domain = self.parameters.get("Domain", "Default")
        user = self.parameters.get("User")
        password = self.parameters.get("Password")
        appcred_file = self.parameters.get("Appcred")
        authDict = {}
        authArgs = {}
        if user and password:
            authDict = {
                "auth": {
                    "identity": {
                        "methods": ["password"],
                        "password": {"user": {"name": user, "domain": {"name": domain}, "password": password}},
                    }
                }
            }
        elif self.parameters.get("Auth") == "voms":
            authDict = {
                "auth": {
                    "identity": {
                        "methods": ["mapped"],
                        "mapped": {"voms": True, "identity_provider": "egi.eu", "protocol": "mapped"},
                    }
                }
            }
            if self.parameters.get("Proxy"):
                authArgs["cert"] = self.parameters.get("Proxy")
        elif appcred_file:
            # The application credentials are stored in a file of the format:
            # id secret
            ac_fd = open(appcred_file, "r")
            auth_info = ac_fd.read()
            auth_info = auth_info.strip()
            ac_id, ac_secret = auth_info.split(" ", 1)
            ac_fd.close()
            authDict = {
                "auth": {
                    "identity": {
                        "methods": ["application_credential"],
                        "application_credential": {"id": ac_id, "secret": ac_secret},
                    }
                }
            }
        else:
            return S_ERROR("No valid credentials provided")

        # appcred includes the project scope binding in the credential itself
        if self.project and not appcred_file:
            authDict["auth"]["scope"] = {"project": {"domain": {"name": domain}, "name": self.project}}

        gLogger.debug("Request token with auth arguments: %s and body %s" % (str(authArgs), str(authDict)))

        url = "%s/auth/tokens" % self.url
        try:
            result = requests.post(
                url,
                headers={
                    "Content-Type": "application/json",
                    "Accept": "application/json",
                },
                json=authDict,
                verify=self.caPath,
                **authArgs
            )

        except Exception as exc:
            return S_ERROR("Exception getting keystone token: %s" % str(exc))

        if result.status_code not in [200, 201, 202, 203, 204]:
            return S_ERROR("Failed to get keystone token: %s" % result.text)

        try:
            self.token = result.headers["X-Subject-Token"]
        except Exception as exc:
            return S_ERROR("Failed to get keystone token: %s" % str(exc))

        output = result.json()

        expires = fromString(str(output["token"]["expires_at"]).replace("T", " ").replace("Z", ""))
        issued = fromString(str(output["token"]["issued_at"]).replace("T", " ").replace("Z", ""))
        self.expires = dateTime() + (expires - issued)

        if "project" in output["token"]:
            if output["token"]["project"]["name"] == self.project:
                self.projectID = output["token"]["project"]["id"]

        if "catalog" in output["token"]:
            for service in output["token"]["catalog"]:
                if service["type"] == "compute":
                    for endpoint in service["endpoints"]:
                        if endpoint["interface"] == "public":
                            self.computeURL = str(endpoint["url"])

                elif service["type"] == "image":
                    for endpoint in service["endpoints"]:
                        if endpoint["interface"] == "public":
                            self.imageURL = str(endpoint["url"])

                elif service["type"] == "network":
                    for endpoint in service["endpoints"]:
                        if endpoint["interface"] == "public":
                            self.networkURL = str(endpoint["url"])

        return S_OK(self.token)

    def getTenants(self):
        """Get available tenants for the current token

        :return: S_OK((tenant, tenant_id)) or S_ERROR
        """

        if self.token is None:
            return S_ERROR("No Keystone token yet available")

        try:
            result = requests.get(
                "%s/tenants" % self.url,
                headers={"Content-Type": "application/json", "X-Auth-Token": self.token},
                verify=self.caPath,
            )
        except Exception as exc:
            return S_ERROR("Failed to get keystone token: %s" % str(exc))

        if result.status_code != 200:
            return S_ERROR("Error: %s" % result.text)

        output = result.json()
        tenants = []
        if "tenants" in output:
            for item in output["tenants"]:
                tenants.append((item["name"], item["id"]))

        return S_OK(tenants)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-list-users
# Author :  Adrian Casajus
########################################################################
"""
Lists the users in the Configuration. If no group is specified return all users.

Example:
  $ dirac-admin-list-users
  All users registered:
  vhamar
  msapunov
  atsareg
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerSwitch("e", "extended", "Show extended info")
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["Group:    Only users from this group (default: all)"], default=["all"], mandatory=False)
    Script.parseCommandLine(ignoreErrors=True)
    args = Script.getPositionalArgs(group=True)

    import DIRAC
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()
    exitCode = 0
    errorList = []
    extendedInfo = False

    for unprocSw in Script.getUnprocessedSwitches():
        if unprocSw[0] in ("e", "extended"):
            extendedInfo = True

    def printUsersInGroup(group=False):
        result = diracAdmin.csListUsers(group)
        if result["OK"]:
            if group:
                print("Users in group %s:" % group)
            else:
                print("All users registered:")
            for username in result["Value"]:
                print(" %s" % username)

    def describeUsersInGroup(group=False):
        result = diracAdmin.csListUsers(group)
        if result["OK"]:
            if group:
                print("Users in group %s:" % group)
            else:
                print("All users registered:")
            result = diracAdmin.csDescribeUsers(result["Value"])
            print(diracAdmin.pPrint.pformat(result["Value"]))

    for group in args:
        if "all" in args:
            group = False
        if not extendedInfo:
            printUsersInGroup(group)
        else:
            describeUsersInGroup(group)

    for error in errorList:
        print("ERROR %s: %s" % error)

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" :mod: DErrno

    ==========================

    .. module: DErrno

    :synopsis: Error list and utilities for handling errors in DIRAC


    This module contains list of errors that can be encountered in DIRAC.
    It complements the errno module of python.

    It also contains utilities to manipulate these errors.

    Finally, it contains a DErrno class that contains an error number
    as well as a low level error message. It behaves like a string for
    compatibility reasons

    In order to add extension specific error, you need to create in your extension the file
    Core/Utilities/DErrno.py, which will contain the following dictionary:

      * extra_dErrName: keys are the error name, values the number of it
      * extra_dErrorCode: same as dErrorCode. keys are the error code, values the name
                          (we don't simply revert the previous dict in case we do not
                          have a one to one mapping)
      * extra_dStrError: same as dStrError, Keys are the error code, values the error description
      * extra_compatErrorString: same as compatErrorString. The compatible error strings are
                                 added to the existing one, and not replacing them.


    Example of extension file :

       * extra_dErrName = { 'ELHCBSPE' : 3001 }
       * extra_dErrorCode = { 3001 : 'ELHCBSPE'}
       * extra_dStrError = { 3001 : "This is a description text of the specific LHCb error" }
       * extra_compatErrorString = { 3001 : ["living easy, living free"],
                             DErrno.ERRX : ['An error message for ERRX that is specific to LHCb']}

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import importlib
import sys

from DIRAC.Core.Utilities.Extensions import extensionsByPriority

# To avoid conflict, the error numbers should be greater than 1000
# We decided to group the by range of 100 per system

# 1000: Generic
# 1100: Core
# 1200: Framework
# 1300: Interfaces
# 1400: Config
# 1500: WMS + Workflow
# 1600: DMS + StorageManagement
# 1700: RMS
# 1800: Accounting + Monitoring
# 1900: TS + Production
# 2000: Resources + RSS

# ## Generic (10XX)
# Python related: 0X
ETYPE = 1000
EIMPERR = 1001
ENOMETH = 1002
ECONF = 1003
EVALUE = 1004
EEEXCEPTION = 1005
# Files manipulation: 1X
ECTMPF = 1010
EOF = 1011
ERF = 1012
EWF = 1013
ESPF = 1014

# ## Core (11XX)
# Certificates and Proxy: 0X
EX509 = 1100
EPROXYFIND = 1101
EPROXYREAD = 1102
ECERTFIND = 1103
ECERTREAD = 1104
ENOCERT = 1105
ENOCHAIN = 1106
ENOPKEY = 1107
ENOGROUP = 1108
# DISET: 1X
EDISET = 1110
ENOAUTH = 1111
# 3rd party security: 2X
E3RDPARTY = 1120
EVOMS = 1121
# Databases : 3X
EDB = 1130
EMYSQL = 1131
ESQLA = 1132
# Message Queues: 4X
EMQUKN = 1140
EMQNOM = 1141
EMQCONN = 1142
# Elasticsearch
EELNOFOUND = 1146

# config
ESECTION = 1400

# processes
EEZOMBIE = 1147
EENOPID = 1148

# ## WMS/Workflow
EWMSUKN = 1500
EWMSJDL = 1501
EWMSRESC = 1502
EWMSSUBM = 1503
EWMSJMAN = 1504
EWMSSTATUS = 1505
EWMSNOPILOT = 1550

# ## DMS/StorageManagement (16XX)
EFILESIZE = 1601
EGFAL = 1602
EBADCKS = 1603
EFCERR = 1604

# ## RMS (17XX)
ERMSUKN = 1700

# ## TS (19XX)
ETSUKN = 1900
ETSDATA = 1901

# ## Resources and RSS (20XX)
ERESGEN = 2000
ERESUNA = 2001
ERESUNK = 2002

# This translates the integer number into the name of the variable
dErrorCode = {
    # ## Generic (10XX)
    # 100X: Python related
    1000: "ETYPE",
    1001: "EIMPERR",
    1002: "ENOMETH",
    1003: "ECONF",
    1004: "EVALUE",
    1005: "EEEXCEPTION",
    # 101X: Files manipulation
    1010: "ECTMPF",
    1011: "EOF",
    1012: "ERF",
    1013: "EWF",
    1014: "ESPF",
    # ## Core
    # 110X: Certificates and Proxy
    1100: "EX509",
    1101: "EPROXYFIND",
    1102: "EPROXYREAD",
    1103: "ECERTFIND",
    1104: "ECERTREAD",
    1105: "ENOCERT",
    1106: "ENOCHAIN",
    1107: "ENOPKEY",
    1108: "ENOGROUP",
    # 111X: DISET
    1110: "EDISET",
    1111: "ENOAUTH",
    # 112X: 3rd party security
    1120: "E3RDPARTY",
    1121: "EVOMS",
    # 113X: Databases
    1130: "EDB",
    1131: "EMYSQL",
    1132: "ESQLA",
    # 114X: Message Queues
    1140: "EMQUKN",
    1141: "EMQNOM",
    1142: "EMQCONN",
    # Elasticsearch
    1146: "EELNOFOUND",
    # Config
    1400: "ESECTION",
    # Processes
    1147: "EEZOMBIE",
    1148: "EENOPID",
    # WMS/Workflow
    1500: "EWMSUKN",
    1501: "EWMSJDL",
    1502: "EWMSRESC",
    1503: "EWMSSUBM",
    1504: "EWMSJMAN",
    1505: "EWMSSTATUS",
    1550: "EWMSNOPILOT",
    # DMS/StorageManagement
    1601: "EFILESIZE",
    1602: "EGFAL",
    1603: "EBADCKS",
    1604: "EFCERR",
    # RMS
    1700: "ERMSUKN",
    # Resources and RSS
    2000: "ERESGEN",
    2001: "ERESUNA",
    2002: "ERESUNK",
    # TS
    1900: "ETSUKN",
    1901: "ETSDATA",
}


dStrError = {  # Generic (10XX)
    # 100X: Python related
    ETYPE: "Object Type Error",
    EIMPERR: "Failed to import library",
    ENOMETH: "No such method or function",
    ECONF: "Configuration error",
    EVALUE: "Wrong value passed",
    EEEXCEPTION: "runtime general exception",
    # 101X: Files manipulation
    ECTMPF: "Failed to create temporary file",
    EOF: "Cannot open file",
    ERF: "Cannot read from file",
    EWF: "Cannot write to file",
    ESPF: "Cannot set permissions to file",
    # ## Core
    # 110X: Certificates and Proxy
    EX509: "Generic Error with X509",
    EPROXYFIND: "Can't find proxy",
    EPROXYREAD: "Can't read proxy",
    ECERTFIND: "Can't find certificate",
    ECERTREAD: "Can't read certificate",
    ENOCERT: "No certificate loaded",
    ENOCHAIN: "No chain loaded",
    ENOPKEY: "No private key loaded",
    ENOGROUP: "No DIRAC group",
    # 111X: DISET
    EDISET: "DISET Error",
    ENOAUTH: "Unauthorized query",
    # 112X: 3rd party security
    E3RDPARTY: "3rd party security service error",
    EVOMS: "VOMS Error",
    # 113X: Databases
    EDB: "Database Error",
    EMYSQL: "MySQL Error",
    ESQLA: "SQLAlchemy Error",
    # 114X: Message Queues
    EMQUKN: "Unknown MQ Error",
    EMQNOM: "No messages",
    EMQCONN: "MQ connection failure",
    # 114X Elasticsearch
    EELNOFOUND: "Index not found",
    # Config
    ESECTION: "Section is not found",
    # processes
    EEZOMBIE: "Zombie process",
    EENOPID: "No PID of process",
    # WMS/Workflow
    EWMSUKN: "Unknown WMS error",
    EWMSJDL: "Invalid job description",
    EWMSRESC: "Job to reschedule",
    EWMSSUBM: "Job submission error",
    EWMSJMAN: "Job management error",
    EWMSSTATUS: "Job status error",
    EWMSNOPILOT: "No pilots found",
    # DMS/StorageManagement
    EFILESIZE: "Bad file size",
    EGFAL: "Error with the gfal call",
    EBADCKS: "Bad checksum",
    EFCERR: "FileCatalog error",
    # RMS
    ERMSUKN: "Unknown RMS error",
    # Resources and RSS
    ERESGEN: "Unknown Resource Failure",
    ERESUNA: "Resource not available",
    ERESUNK: "Unknown Resource",
    # TS
    ETSUKN: "Unknown Transformation System Error",
    ETSDATA: "Invalid Input Data definition",
}


def strerror(code):
    """This method wraps up os.strerror, and behave the same way.
    It completes it with the DIRAC specific errors.
    """

    if code == 0:
        return "Undefined error"

    errMsg = "Unknown error %s" % code

    try:
        errMsg = dStrError[code]
    except KeyError:
        # It is not a DIRAC specific error, try the os one
        try:
            errMsg = os.strerror(code)
            # On some system, os.strerror raises an exception with unknown code,
            # on others, it returns a message...
        except ValueError:
            pass

    return errMsg


def cmpError(inErr, candidate):
    """This function compares an error (in its old form (a string or dictionary) or in its int form
    with a candidate error code.

    :param inErr: a string, an integer, a S_ERROR dictionary
    :type inErr: str or int or S_ERROR
    :param int candidate: error code to compare with

    :return: True or False

    If an S_ERROR instance is passed, we compare the code with S_ERROR['Errno']
    If it is a Integer, we do a direct comparison
    If it is a String, we use strerror to check the error string
    """

    if isinstance(inErr, six.string_types):  # old style
        # Compare error message strings
        errMsg = strerror(candidate)
        return errMsg in inErr
    elif isinstance(inErr, dict):  # if the S_ERROR structure is given
        # Check if Errno defined in the dict
        errorNumber = inErr.get("Errno")
        if errorNumber:
            return errorNumber == candidate
        errMsg = strerror(candidate)
        return errMsg in inErr.get("Message", "")
    elif isinstance(inErr, int):
        return inErr == candidate
    else:
        raise TypeError("Unknown input error type %s" % type(inErr))


def includeExtensionErrors():
    """Merge all the errors of all the extensions into the errors of these modules
    Should be called only at the initialization of DIRAC, so by the parseCommandLine,
    dirac-agent.py, dirac-service.py, dirac-executor.py
    """
    for extension in reversed(extensionsByPriority()):
        if extension == "DIRAC":
            continue
        try:
            ext_derrno = importlib.import_module("%s.Core.Utilities.DErrno" % extension)
        except ImportError:
            pass
        else:
            # The next 3 dictionary MUST be present for consistency
            # Global name of errors
            sys.modules[__name__].__dict__.update(ext_derrno.extra_dErrName)
            # Dictionary with the error codes
            sys.modules[__name__].dErrorCode.update(ext_derrno.extra_dErrorCode)
            # Error description string
            sys.modules[__name__].dStrError.update(ext_derrno.extra_dStrError)

            # extra_compatErrorString is optional
            for err in getattr(ext_derrno, "extra_compatErrorString", []):
                sys.modules[__name__].compatErrorString.setdefault(err, []).extend(
                    ext_derrno.extra_compatErrorString[err]
                )



""" ResourceStatusHandler

  Module that allows users to access the ResourceStatusDB remotely.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=too-many-arguments

__RCSID__ = "$Id$"

import six

from DIRAC import gLogger, S_OK
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader


def convert(table, params):
    """Conversion utility for backward compatibility"""
    gLogger.debug("Calls from old client")
    # In this case, "params" contain the "meta", so we at least need to swap!
    tableFromOldCall = params["table"]
    columnsFromOldCall = params.get("columns")
    olderFromOldCall = params.get("older")
    newerFromOldCall = params.get("newer")
    orderFromOldCall = params.get("order")
    limitFromOldCall = params.get("limit")
    params = table
    if columnsFromOldCall or olderFromOldCall or newerFromOldCall or orderFromOldCall or limitFromOldCall:
        params["Meta"] = {}
        if columnsFromOldCall:
            params["Meta"]["columns"] = columnsFromOldCall
        if olderFromOldCall:
            params["Meta"]["older"] = olderFromOldCall
        if newerFromOldCall:
            params["Meta"]["newer"] = newerFromOldCall
        if orderFromOldCall:
            params["Meta"]["order"] = orderFromOldCall
        if limitFromOldCall:
            params["Meta"]["limit"] = limitFromOldCall
    table = tableFromOldCall

    return params, table


def loadResourceStatusComponent(moduleName, className):
    """
    Create an object of a given database component.

    :param moduleName: module name to be loaded
    :param className: class name
    :return: object instance wrapped in a standard Dirac return object.
    """

    objectLoader = ObjectLoader()
    componentModule = "ResourceStatusSystem.DB.%s" % (moduleName,)
    result = objectLoader.loadObject(componentModule, className)
    if not result["OK"]:
        gLogger.error("Failed to load RSS component", "%s: %s" % (moduleName, result["Message"]))
        return result
    componentClass = result["Value"]
    component = componentClass()
    return S_OK(component)


class ResourceStatusHandler(RequestHandler):
    """
    The ResourceStatusHandler exposes the DB front-end functions through a XML-RPC
    server, functionalities inherited from
    :class:`DIRAC.Core.DISET.RequestHandler.RequestHandler`

    According to the ResourceStatusDB philosophy, only functions of the type:
    - insert
    - update
    - select
    - delete

    are exposed. If you need anything more complicated, either look for it on the
    :class:`ResourceStatusClient`, or code it yourself. This way the DB and the
    Service are kept clean and tidied.

    To can use this service on this way, but you MUST NOT DO IT. Use it through the
    :class:`ResourceStatusClient`. If offers in the worst case as good performance
    as the :class:`ResourceStatusHandler`, if not better.

     >>> from DIRAC.Core.DISET.RPCClient import RPCCLient
     >>> server = RPCCLient( "ResourceStatus/ResourceStatus" )
    """

    def __init__(self, *args, **kwargs):

        super(ResourceStatusHandler, self).__init__(*args, **kwargs)

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """
        Handler initialization, where we:
          dynamically load ResourceStatus database plugin module, as advised by the config,
          (assumes that the module name and a class name are the same)
          set the ResourceManagementDB as global db.

          :param serviceInfoDict: service info dictionary
          :return: standard Dirac return object

        """

        defaultOption, defaultClass = "ResourceStatusDB", "ResourceStatusDB"
        configValue = getServiceOption(serviceInfoDict, defaultOption, defaultClass)
        result = loadResourceStatusComponent(configValue, configValue)

        if not result["OK"]:
            return result

        cls.db = result["Value"]

        return S_OK()

    def __logResult(self, methodName, result):
        """
        Method that writes to log error messages
        """

        if not result["OK"]:
            self.log.error("%s%s" % (methodName, result["Message"]))

    types_insert = [[six.string_types, dict], dict]

    def export_insert(self, table, params):
        """
        This method is a bridge to access :class:`ResourceStatusDB` remotely. It
        does not add neither processing nor validation. If you need to know more
        about this method, you must keep reading on the database documentation.

        :Parameters:
          **table** - `string` or `dict`
            should contain the table from which querying
            if it's a `dict` the query comes from a client prior to v6r18

          **params** - `dict`
            arguments for the mysql query. Currently it is being used only for column selection.
            For example: meta = { 'columns' : [ 'Name' ] } will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """

        self.log.info("insert: %s %s" % (table, params))
        res = self.db.insert(table, params)
        self.__logResult("insert", res)

        return res

    types_select = [[six.string_types, dict], dict]

    def export_select(self, table, params):
        """
        This method is a bridge to access :class:`ResourceStatusDB` remotely. It
        does not add neither processing nor validation. If you need to know more
        about this method, you must keep reading on the database documentation.

        :Parameters:
          **table** - `string` or `dict`
            should contain the table from which querying
            if it's a `dict` the query comes from a client prior to v6r18

          **params** - `dict`
            arguments for the mysql query. Currently it is being used only for column selection.
            For example: meta = { 'columns' : [ 'Name' ] } will return only the 'Name' column.

        :return: S_OK() || S_ERROR()
        """

        self.log.info("select: %s %s" % (table, params))
        res = self.db.select(table, params)
        self.__logResult("select", res)

        return res

    types_delete = [[six.string_types, dict], dict]

    def export_delete(self, table, params):
        """
    This method is a bridge to access :class:`ResourceStatusDB` remotely.\
    It does not add neither processing nor validation. If you need to know more \
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **table** - `string` or `dict`
        should contain the table from which querying
        if it's a `dict` the query comes from a client prior to v6r18

      **params** - `dict`
        arguments for the mysql query. Currently it is being used only for column selection.
        For example: meta = { 'columns' : [ 'Name' ] } will return only the 'Name' column.


    :return: S_OK() || S_ERROR()
    """

        self.log.info("delete: %s %s" % (table, params))
        res = self.db.delete(table, params)
        self.__logResult("delete", res)

        return res

    types_addOrModify = [[six.string_types, dict], dict]

    def export_addOrModify(self, table, params):
        """
    This method is a bridge to access :class:`ResourceStatusDB` remotely.\
    It does not add neither processing nor validation. If you need to know more \
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **table** - `string` or `dict`
        should contain the table from which querying
        if it's a `dict` the query comes from a client prior to v6r18

      **params** - `dict`
        arguments for the mysql query. Currently it is being used only for column selection.
        For example: meta = { 'columns' : [ 'Name' ] } will return only the 'Name' column.


    :return: S_OK() || S_ERROR()
    """

        self.log.info("addOrModify: %s %s" % (table, params))
        res = self.db.addOrModify(table, params)
        self.__logResult("addOrModify", res)

        return res

    types_addIfNotThere = [[six.string_types, dict], dict]

    def export_addIfNotThere(self, table, params):
        """
    This method is a bridge to access :class:`ResourceStatusDB` remotely.\
    It does not add neither processing nor validation. If you need to know more \
    about this method, you must keep reading on the database documentation.

    :Parameters:
      **table** - `string` or `dict`
        should contain the table from which querying
        if it's a `dict` the query comes from a client prior to v6r18

      **params** - `dict`
        arguments for the mysql query. Currently it is being used only for column selection.
        For example: meta = { 'columns' : [ 'Name' ] } will return only the 'Name' column.


    :return: S_OK() || S_ERROR()
    """

        self.log.info("addIfNotThere: %s %s" % (table, params))
        res = self.db.addIfNotThere(table, params)
        self.__logResult("addIfNotThere", res)

        return res



#!/usr/bin/env python
########################################################################
# File :   dirac-agent
# Author : Adria Casajus, Andrei Tsaregorodtsev, Stuart Paterson
########################################################################
"""
This is a script to launch DIRAC agents. Mostly internal.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import sys

from DIRAC import gLogger
from DIRAC.Core.Base.AgentReactor import AgentReactor
from DIRAC.Core.Utilities.DErrno import includeExtensionErrors
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.registerArgument(["Agent: specify which agent to run"])
    positionalArgs = Script.getPositionalArgs(group=True)
    localCfg = Script.localCfg

    agentName = positionalArgs[0]
    localCfg.setConfigurationForAgent(agentName)
    localCfg.addMandatoryEntry("/DIRAC/Setup")
    localCfg.addDefaultEntry("/DIRAC/Security/UseServerCertificate", "yes")
    localCfg.addDefaultEntry("LogLevel", "INFO")
    localCfg.addDefaultEntry("LogColor", True)
    resultDict = localCfg.loadUserData()
    if not resultDict["OK"]:
        gLogger.error("There were errors when loading configuration", resultDict["Message"])
        sys.exit(1)

    includeExtensionErrors()

    agentReactor = AgentReactor(positionalArgs[0])
    result = agentReactor.loadAgentModules(positionalArgs)
    if result["OK"]:
        agentReactor.go()
    else:
        gLogger.error("Error while loading agent module", result["Message"])
        sys.exit(2)


if __name__ == "__main__":
    main()



"""
Profiling class for updated information on process status
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import datetime
import errno
import psutil

from DIRAC import gLogger, S_OK, S_ERROR
from DIRAC.Core.Utilities.DErrno import EEZOMBIE, EENOPID, EEEXCEPTION


def checkInvocation(func):
    """Decorator for invoking psutil methods"""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except psutil.ZombieProcess as e:
            gLogger.error("Zombie process: %s" % e)
            return S_ERROR(EEZOMBIE, "Zombie process: %s" % e)
        except psutil.NoSuchProcess as e:
            gLogger.error("No such process: %s" % e)
            return S_ERROR(errno.ESRCH, "No such process: %s" % e)
        except psutil.AccessDenied as e:
            gLogger.error("Access denied: %s" % e)
            return S_ERROR(errno.EPERM, "Access denied: %s" % e)
        except Exception as e:  # pylint: disable=broad-except
            gLogger.error(e)
            return S_ERROR(EEEXCEPTION, e)

    return wrapper


class Profiler(object):
    """
    Class for profiling both general stats about a machine and individual processes.
    Every instance of this class is associated to a single process by using its PID.
    Calls to the different methods of the class will return the current state of the process.
    """

    def __init__(self, pid=None):
        """
        :param str pid: PID of the process to be profiled
        """
        self.process = None
        if pid:
            try:
                self.process = psutil.Process(int(pid))
            except psutil.NoSuchProcess as e:
                gLogger.error("No such process: %s" % e)

    def pid(self):
        """
        Returns the process PID
        """
        if self.process:
            return S_OK(self.process.pid)
        else:
            gLogger.error("No PID of process to profile")
            return S_ERROR(EENOPID, "No PID of process to profile")

    @checkInvocation
    def status(self):
        """Returns the process status"""
        result = self.process.status()
        return S_OK(result)

    @checkInvocation
    def runningTime(self):
        """
        Returns the uptime of the process
        """
        start = datetime.datetime.fromtimestamp(self.process.create_time())
        result = (datetime.datetime.now() - start).total_seconds()
        return S_OK(result)

    @checkInvocation
    def memoryUsage(self, withChildren=False):
        """
        Returns the memory usage of the process in MB
        """
        # Information is returned in bytes
        rss = self.process.memory_info().rss
        if withChildren:
            for child in self.process.children(recursive=True):
                rss += child.memory_info().rss
        # converted to MB
        return S_OK(rss / float(2 ** 20))

    @checkInvocation
    def vSizeUsage(self, withChildren=False):
        """
        Returns the memory usage of the process in MB
        """
        # Information is returned in bytes
        vms = self.process.memory_info().vms
        if withChildren:
            for child in self.process.children(recursive=True):
                vms += child.memory_info().vms
        # converted to MB
        return S_OK(vms / float(2 ** 20))

    @checkInvocation
    def numThreads(self, withChildren=False):
        """
        Returns the number of threads the process is using
        """
        nThreads = self.process.num_threads()
        if withChildren:
            for child in self.process.children(recursive=True):
                nThreads += child.num_threads()
        return S_OK(nThreads)

    @checkInvocation
    def cpuPercentage(self, withChildren=False):
        """
        Returns the percentage of cpu used by the process
        """
        cpuPercentage = self.process.cpu_percent()
        if withChildren:
            for child in self.process.children(recursive=True):
                cpuPercentage += child.cpu_percent()
        return S_OK(cpuPercentage)

    @checkInvocation
    def cpuUsageUser(self, withChildren=False, withTerminatedChildren=False):
        """
        Returns the percentage of cpu used by the process
        """
        cpuUsageUser = self.process.cpu_times().user
        childrenUser = 0
        oldChildrenUser = 0
        if withChildren:  # active children
            for child in self.process.children(recursive=True):
                childrenUser += child.cpu_times().user
            gLogger.debug("CPU user (process, children)", "(%.1fs, %.1fs)" % (cpuUsageUser, childrenUser))
        if withTerminatedChildren:  # all terminated children of the root process
            for child in self.process.children(recursive=True):
                oldChildrenUser += child.cpu_times().children_user
            gLogger.debug("CPU user (process, old children)", "(%.1fs, %.1fs)" % (cpuUsageUser, oldChildrenUser))
        else:
            gLogger.debug("CPU user", "%.1fs" % cpuUsageUser)
        return S_OK(cpuUsageUser + childrenUser + oldChildrenUser)

    @checkInvocation
    def cpuUsageSystem(self, withChildren=False, withTerminatedChildren=False):
        """
        Returns the percentage of cpu used by the process
        """
        cpuUsageSystem = self.process.cpu_times().system
        childrenSystem = 0
        oldChildrenSystem = 0
        if withChildren:  # active children
            for child in self.process.children(recursive=True):
                childrenSystem += child.cpu_times().system
            gLogger.debug("CPU user (process, children)", "(%.1fs, %.1fs)" % (cpuUsageSystem, childrenSystem))
        if withTerminatedChildren:  # all terminated children of the root process
            for child in self.process.children(recursive=True):
                oldChildrenSystem += child.cpu_times().children_system
            gLogger.debug("CPU user (process, old children)", "(%.1fs, %.1fs)" % (cpuUsageSystem, oldChildrenSystem))
        else:
            gLogger.debug("CPU user", "%.1fs" % cpuUsageSystem)
        return S_OK(cpuUsageSystem + childrenSystem + oldChildrenSystem)

    def getAllProcessData(self, withChildren=False, withTerminatedChildren=False):
        """
        Returns data available about a process
        """
        data = {}

        data["datetime"] = datetime.datetime.utcnow()
        data["stats"] = {}

        result = self.pid()
        if result["OK"]:
            data["stats"]["pid"] = result["Value"]

        result = self.status()
        if result["OK"]:
            data["stats"]["status"] = result["Value"]

        result = self.runningTime()
        if result["OK"]:
            data["stats"]["runningTime"] = result["Value"]

        result = self.memoryUsage(withChildren)
        if result["OK"]:
            data["stats"]["memoryUsage"] = result["Value"]

        result = self.vSizeUsage(withChildren)
        if result["OK"]:
            data["stats"]["vSizeUsage"] = result["Value"]

        result = self.numThreads(withChildren)
        if result["OK"]:
            data["stats"]["threads"] = result["Value"]

        result = self.cpuPercentage(withChildren)
        if result["OK"]:
            data["stats"]["cpuPercentage"] = result["Value"]

        result = self.cpuUsageUser(withChildren, withTerminatedChildren)
        if result["OK"]:
            data["stats"]["cpuUsageUser"] = result["Value"]

        result = self.cpuUsageSystem(withChildren, withTerminatedChildren)
        if result["OK"]:
            data["stats"]["cpuUsageSystem"] = result["Value"]

        return S_OK(data)



""" This is used to cache already generated plots in order to improve the
    response time and increase reusability.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import hashlib
import time
import threading

from DIRAC import S_OK, gLogger
from DIRAC.Core.Utilities.ThreadSafe import Synchronizer
from DIRAC.FrameworkSystem.private.monitoring.RRDManager import RRDManager
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor


gSynchro = Synchronizer()


class PlotCache(object):
    """
    This is generally used for caching the plots.
    """

    # This class is basically used to cache the graphs that are generated once so if the user tries to generate them again
    # the entire process of graph generation using RRDManager is not performed instead a cached version of the graph is
    # returned to reduce the response time.

    def __init__(self, rrdManager=None):
        if rrdManager is None:
            self.rrdManager = RRDManager
        else:
            self.rrdManager = rrdManager
        self.plotsLocation = self.rrdManager.getGraphLocation()
        for plot in os.listdir(self.plotsLocation):
            if plot.find(".png") > 0:
                os.unlink("%s/%s" % (self.plotsLocation, plot))
        self.cachedPlots = {}
        self.alive = True
        self.graceTime = 60
        self.purgeThread = threading.Thread(target=self.purgeCached)
        self.purgeThread.start()

    def __generateName(self, *args, **kwargs):
        m = hashlib.md5()
        m.update(repr(args).encode())
        m.update(repr(kwargs).encode())
        return m.hexdigest()

    def __isCurrentTime(self, toSecs):
        currentBucket = self.rrdManager.getCurrentBucketTime(self.graceTime)
        return toSecs + self.graceTime > currentBucket

    def purgeCached(self):
        while self.alive:
            time.sleep(self.graceTime * 2)
            self.__purgeExpiredGraphs()

    @gSynchro
    def __refreshGraph(self, graphFile):
        self.cachedPlots[graphFile][1] = time.time()

    @gSynchro
    def __registerGraph(self, graphFile, fromSecs, toSecs):
        if self.__isCurrentTime(toSecs):
            cacheTime = self.graceTime * 2
        else:
            cacheTime = 3600
        self.cachedPlots[graphFile] = [cacheTime, time.time()]

    @gSynchro
    def __isCacheGraph(self, graphFile):
        return graphFile in self.cachedPlots

    @gSynchro
    def __purgeExpiredGraphs(self):
        now = time.time()
        graphsToDelete = []
        for cachedFile in self.cachedPlots:
            fileData = self.cachedPlots[cachedFile]
            if fileData[0] + fileData[1] < now:
                graphsToDelete.append(cachedFile)
        for cachedFile in graphsToDelete:
            try:
                filePath = "%s/%s" % (self.plotsLocation, cachedFile)
                os.unlink(filePath)
            except Exception as e:
                gLogger.error("Can't delete plot file", "%s: %s" % (filePath, str(e)))
            del self.cachedPlots[cachedFile]

    def groupPlot(self, *args):
        """
        This method is called by the __generateGroupPlots method of the ServiceInterface in order to cache
        the generated group plots.
        """
        return self.__getGraph(getattr(self.rrdManager, "groupPlot"), args)

    def plot(self, *args):
        """
        This method is called by the __generatePlots method of the ServiceInterface in order to cache
        the generated plot.
        """
        return self.__getGraph(getattr(self.rrdManager, "plot"), args)

    def __getGraph(self, plotFunc, args):
        fromSecs = args[0]
        toSecs = args[1]
        graphFile = "%s-%s-%s.png" % (
            self.__generateName(*args[2:]),
            self.rrdManager.bucketize(fromSecs, self.graceTime),
            self.rrdManager.bucketize(toSecs, self.graceTime),
        )
        if self.__isCacheGraph(graphFile):
            self.__refreshGraph(graphFile)
            gLogger.info("Cached graph file %s" % graphFile)
            gMonitor.addMark("cachedplots")
            return S_OK(graphFile)
        else:
            gMonitor.addMark("drawnplots")
            self.__registerGraph(graphFile, fromSecs, toSecs)
            return plotFunc(graphFilename=graphFile, *args)



#!/usr/bin/env python
"""
Get parameters assigned to the CE
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import json
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

ceName = ""
Queue = ""
Site = ""


@Script()
def main():
    global ceName
    global Queue
    global Site

    from DIRAC import gLogger, exit as DIRACExit
    from DIRAC.ConfigurationSystem.Client.Helpers import Resources

    def setCEName(args):
        global ceName
        ceName = args

    def setSite(args):
        global Site
        Site = args

    def setQueue(args):
        global Queue
        Queue = args

    Script.registerSwitch("N:", "Name=", "Computing Element Name (Mandatory)", setCEName)
    Script.registerSwitch("S:", "Site=", "Site Name (Mandatory)", setSite)
    Script.registerSwitch("Q:", "Queue=", "Queue Name (Mandatory)", setQueue)

    Script.parseCommandLine(ignoreErrors=True)

    result = Resources.getQueue(Site, ceName, Queue)

    if not result["OK"]:
        # Normal DIRAC queue search failed, check for matching VM images
        vmresult = Resources.getVMTypeConfig(Site, ceName, Queue)
        if vmresult["OK"]:
            gLogger.notice(json.dumps(vmresult["Value"]))
            return
        # Queue & VM not found, return original queue failure message
        gLogger.error("Could not retrieve resource parameters", ": " + vmresult["Message"])
        DIRACExit(1)
    gLogger.notice(json.dumps(result["Value"]))


if __name__ == "__main__":
    main()



"""
This is the client of the Monitoring service based on Elasticsearch.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities.Plotting.FileCoding import codeRequestInFileId

__RCSID__ = "$Id$"


@createClient("Monitoring/Monitoring")
class MonitoringClient(Client):
    """
    .. class:: MonitoringClient

    This class expose the methods of the Monitoring Service

    """

    def __init__(self, **kwargs):
        """Simple constructor"""

        super(MonitoringClient, self).__init__(**kwargs)
        self.setServer("Monitoring/Monitoring")

    def generateDelayedPlot(
        self, typeName, reportName, startTime, endTime, condDict, grouping, extraArgs=None, compress=True
    ):
        """
        It is used to encode the plot parameters used to create a certain plot.

        :param str typeName: the type of the monitoring
        :param int startTime: epoch time, start time of the plot
        :param int endTime: epoch time, end time of the plot
        :param dict condDict: is the conditions used to gnerate the plot: {'Status':['Running'],'grouping': ['Site'] }
        :param str grouping: is the grouping of the data for example: 'Site'
        :param dict extraArgs: epoch time which can be last day, last week, last month
        :param bool compress: apply compression of the encoded values.

        :return: S_OK(str) or S_ERROR() it returns the encoded plot parameters
        """
        if not isinstance(extraArgs, dict):
            extraArgs = {}
        plotRequest = {
            "typeName": typeName,
            "reportName": reportName,
            "startTime": startTime,
            "endTime": endTime,
            "condDict": condDict,
            "grouping": grouping,
            "extraArgs": extraArgs,
        }
        return codeRequestInFileId(plotRequest, compress)

    def getReport(self, typeName, reportName, startTime, endTime, condDict, grouping, extraArgs=None):
        """
        It is used to get the raw data used to create a plot.

        :param str typeName: the type of the monitoring
        :param str reportName: the name of the plotter used to create the plot for example:  NumberOfJobs
        :param int startTime: epoch time, start time of the plot
        :param int endTime: epoch time, end time of the plot
        :param dict condDict: is the conditions used to gnerate the plot: {'Status':['Running'],'grouping': ['Site'] }
        :param str grouping: is the grouping of the data for example: 'Site'
        :param dict extraArgs: epoch time which can be last day, last week, last month
        :rerturn: S_OK or S_ERROR
        """
        if not isinstance(extraArgs, dict):
            extraArgs = {}
        plotRequest = {
            "typeName": typeName,
            "reportName": reportName,
            "startTime": startTime,
            "endTime": endTime,
            "condDict": condDict,
            "grouping": grouping,
            "extraArgs": extraArgs,
        }
        result = self._getRPC().getReport(plotRequest)
        if "rpcStub" in result:
            del result["rpcStub"]
        return result



""" Class that contains client access to the Publisher handler. """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client


class PublisherClient(Client):
    def __init__(self, **kwargs):

        super(PublisherClient, self).__init__(**kwargs)
        self.setServer("ResourceStatus/Publisher")



""" SingularityCE is a type of "inner" CEs
    (meaning it's used by a jobAgent inside a pilot).
    A computing element class using singularity containers,
    where Singularity is supposed to be found on the WN.

    The goal of this CE is to start the job in the container set by
    the "ContainerRoot" config option.

    DIRAC can be re-installed within the container, extra flags can
    be given to the dirac-install command with the "ContainerExtraOpts"
    option.

    See the Configuration/Resources/Computing documention for details on
    where to set the option parameters.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import io
import json
import os
import shutil
import six
import sys
import tempfile

from six.moves.urllib.request import urlopen

import DIRAC
from DIRAC import S_OK, S_ERROR, gConfig, gLogger
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.ConfigurationSystem.Client.Helpers import CSGlobals
from DIRAC.ConfigurationSystem.Client.Helpers import Operations
from DIRAC.Core.Utilities.ThreadScheduler import gThreadScheduler
from DIRAC.Resources.Computing.ComputingElement import ComputingElement
from DIRAC.WorkloadManagementSystem.Utilities.Utils import createRelocatedJobWrapper

__RCSID__ = "$Id$"

# Default container to use if it isn't specified in the CE options
CONTAINER_DEFROOT = "/cvmfs/cernvm-prod.cern.ch/cvm4"
CONTAINER_WORKDIR = "DIRAC_containers"
CONTAINER_INNERDIR = "/tmp"


# What is executed inside the container (2 options given)

CONTAINER_WRAPPER_INSTALL = """#!/bin/bash

echo "Starting inner container wrapper scripts (uses dirac-install) at `date`."
set -x
cd /tmp
# Avoid using the host's DIRAC(OS) installation
unset DIRAC
unset DIRACOS
# Install DIRAC
./dirac-install.py %(install_args)s
source bashrc
dirac-configure -F %(config_args)s -I
# Add compatibility with pilot3 where config is in pilot.cfg
ln -s etc/dirac.cfg pilot.cfg
# Run next wrapper (to start actual job)
bash %(next_wrapper)s
# Write the payload errorcode to a file for the outer scripts
echo $? > retcode
chmod 644 retcode
echo "Finishing inner container wrapper scripts at `date`."

"""
# Path to a directory on CVMFS to use as a fallback if no
# other version found: Only used if node has user namespaces
FALLBACK_SINGULARITY = "/cvmfs/oasis.opensciencegrid.org/mis/singularity/current/bin"

CONTAINER_WRAPPER_NO_INSTALL = """#!/bin/bash

echo "Starting inner container wrapper scripts (no install) at `date`."
set -x
cd /tmp
export DIRAC=%(dirac_env_var)s
export DIRACOS=%(diracos_env_var)s
# In any case we need to find a bashrc, and a pilot.cfg, both created by the pilot
source %(rc_script)s
# Run next wrapper (to start actual job)
bash %(next_wrapper)s
# Write the payload errorcode to a file for the outer scripts
echo $? > retcode
chmod 644 retcode
echo "Finishing inner container wrapper scripts at `date`."

"""


class SingularityComputingElement(ComputingElement):
    """A Computing Element for running a job within a Singularity container."""

    def __init__(self, ceUniqueID):
        """Standard constructor."""
        super(SingularityComputingElement, self).__init__(ceUniqueID)
        self.__submittedJobs = 0
        self.__runningJobs = 0
        self.__root = CONTAINER_DEFROOT
        if "ContainerRoot" in self.ceParameters:
            self.__root = self.ceParameters["ContainerRoot"]
        self.__workdir = CONTAINER_WORKDIR
        self.__innerdir = CONTAINER_INNERDIR
        self.__singularityBin = "singularity"
        self.__installDIRACInContainer = self.ceParameters.get("InstallDIRACInContainer", six.PY2)
        if isinstance(self.__installDIRACInContainer, six.string_types) and self.__installDIRACInContainer.lower() in (
            "false",
            "no",
        ):
            self.__installDIRACInContainer = False

        self.processors = int(self.ceParameters.get("NumberOfProcessors", 1))

    def __hasUserNS(self):
        """Detect if this node has user namespaces enabled.
        Returns True if they are enabled, False otherwise.
        """
        try:
            with open("/proc/sys/user/max_user_namespaces", "r") as proc_fd:
                maxns = int(proc_fd.readline().strip())
                # Any "reasonable number" of namespaces is sufficient
                return maxns > 100
        except Exception:
            # Any failure, missing file, doesn't contain a number, etc. and we
            # assume they are disabled.
            return False

    def __hasSingularity(self):
        """Search the current PATH for an exectuable named singularity.
        Returns True if it is found, False otherwise.
        """
        if self.ceParameters.get("ContainerBin"):
            binPath = self.ceParameters["ContainerBin"]
            if os.path.isfile(binPath) and os.access(binPath, os.X_OK):
                self.__singularityBin = binPath
                self.log.debug('Use singularity from "%s"' % self.__singularityBin)
                return True
        if "PATH" not in os.environ:
            return False  # Hmm, PATH not set? How unusual...
        searchPaths = os.environ["PATH"].split(os.pathsep)
        # We can use CVMFS as a last resort if userNS is enabled
        if self.__hasUserNS():
            searchPaths.append(FALLBACK_SINGULARITY)
        for searchPath in searchPaths:
            binPath = os.path.join(searchPath, "singularity")
            if os.path.isfile(binPath):
                # File found, check it's executable to be certain:
                if os.access(binPath, os.X_OK):
                    self.log.debug('Found singularity at "%s"' % binPath)
                    self.__singularityBin = binPath
                    return True
        # No suitable binaries found
        return False

    @staticmethod
    def __findInstallBaseDir():
        """Find the path to root of the current DIRAC installation"""
        if six.PY3:
            return os.path.realpath(sys.base_prefix)  # pylint: disable=no-member
        else:
            candidate = os.path.join(DIRAC.rootPath, "bashrc")
            return os.path.dirname(os.path.realpath(candidate))

    def __getInstallFlags(self, infoDict=None):
        """Get the flags to pass to dirac-install.py inside the container.
        Returns a string containing the command line flags.
        """
        if not infoDict:
            infoDict = {}

        instOpts = []

        setup = infoDict.get("DefaultSetup")
        if not setup:
            setup = list(infoDict.get("Setups"))[0]
        if not setup:
            setup = gConfig.getValue("/DIRAC/Setup", "unknown")
        setup = str(setup)

        installationName = str(infoDict.get("Installation"))
        if not installationName or installationName == "None":
            installationName = Operations.Operations(setup=setup).getValue("Pilot/Installation", "")
        if installationName:
            instOpts.append("-V %s" % installationName)

        diracVersions = str(infoDict["Setups"][setup].get("Version")).split(",")
        if not diracVersions:
            diracVersions = str(infoDict["Setups"]["Defaults"].get("Version")).split(",")
        if not diracVersions:
            diracVersions = Operations.Operations(setup=setup).getValue("Pilot/Version", [])
        instOpts.append("-r '%s'" % diracVersions[0].strip())

        pilotExtensionsList = str(infoDict["Setups"][setup].get("CommandExtensions")).split(",")
        if not pilotExtensionsList:
            pilotExtensionsList = str(infoDict["Setups"]["Defaults"].get("CommandExtensions")).split(",")
        if not pilotExtensionsList:
            pilotExtensionsList = Operations.Operations(setup=setup).getValue("Pilot/Extensions", [])
        extensionsList = []
        if pilotExtensionsList:
            if pilotExtensionsList[0] != "None":
                extensionsList = pilotExtensionsList
        else:
            extensionsList = CSGlobals.getCSExtensions()
        if extensionsList:
            instOpts.append("-e '%s'" % ",".join([ext for ext in extensionsList if "Web" not in ext]))
        if "ContainerExtraOpts" in self.ceParameters:
            instOpts.append(self.ceParameters["ContainerExtraOpts"])
        return " ".join(instOpts)

    @staticmethod
    def __getConfigFlags(infoDict=None):
        """Get the flags for dirac-configure inside the container.
        Returns a string containing the command line flags.
        """
        if not infoDict:
            infoDict = {}

        cfgOpts = []

        setup = infoDict.get("DefaultSetup")
        if not setup:
            setup = gConfig.getValue("/DIRAC/Setup", "unknown")
        cfgOpts.append("-S '%s'" % setup)

        csServers = infoDict.get("ConfigurationServers")
        if not csServers:
            csServers = gConfig.getValue("/DIRAC/Configuration/Servers", [])
        cfgOpts.append("-C '%s'" % ",".join([str(ce) for ce in csServers]))
        cfgOpts.append("-n '%s'" % DIRAC.siteName())
        return " ".join(cfgOpts)

    def __createWorkArea(self, jobDesc=None, log=None, logLevel="INFO", proxy=None):
        """Creates a directory for the container and populates it with the
        template directories, scripts & proxy.
        """
        if not jobDesc:
            jobDesc = {}
        if not log:
            log = gLogger

        # Create the directory for our container area
        try:
            os.mkdir(self.__workdir)
        except OSError:
            if not os.path.isdir(self.__workdir):
                result = S_ERROR("Failed to create container base directory '%s'" % self.__workdir)
                result["ReschedulePayload"] = True
                return result
            # Otherwise, directory probably just already exists...
        baseDir = None
        try:
            baseDir = tempfile.mkdtemp(prefix="job%s_" % jobDesc.get("jobID", 0), dir=self.__workdir)
        except OSError:
            result = S_ERROR("Failed to create container work directory in '%s'" % self.__workdir)
            result["ReschedulePayload"] = True
            return result

        self.log.debug("Use singularity workarea: %s" % baseDir)
        for subdir in ["home", "tmp", "var_tmp"]:
            os.mkdir(os.path.join(baseDir, subdir))
        tmpDir = os.path.join(baseDir, "tmp")

        # Now we have a directory, we can stage in the proxy and scripts
        # Proxy
        if proxy:
            proxyLoc = os.path.join(tmpDir, "proxy")
            rawfd = os.open(proxyLoc, os.O_WRONLY | os.O_CREAT, 0o600)
            fd = os.fdopen(rawfd, "wb")
            fd.write(proxy)
            fd.close()
        else:
            self.log.warn("No user proxy")

        # Job Wrapper (Standard-ish DIRAC wrapper)
        result = createRelocatedJobWrapper(
            wrapperPath=tmpDir,
            rootLocation=self.__innerdir,
            jobID=jobDesc.get("jobID", 0),
            jobParams=jobDesc.get("jobParams", {}),
            resourceParams=jobDesc.get("resourceParams", {}),
            optimizerParams=jobDesc.get("optimizerParams", {}),
            log=log,
            logLevel=logLevel,
            extraOptions="" if self.__installDIRACInContainer else "/tmp/pilot.cfg",
        )
        if not result["OK"]:
            result["ReschedulePayload"] = True
            return result
        wrapperPath = result["Value"]

        if self.__installDIRACInContainer:
            if six.PY3:
                result = S_ERROR("InstallDIRACInContainer is not supported with Python 3")
                result["ReschedulePayload"] = True
                return result
            # dirac-install.py

            # Download dirac-install.py
            response = urlopen("https://raw.githubusercontent.com/DIRACGrid/management/master/dirac-install.py")
            code = response.getcode()
            if code > 200 or code >= 300:
                return S_ERROR("Failed to download dirac-install.py with code %s" % code)
            with open("dirac-install.py", "wb") as fp:
                fp.write(response.read())

            install_loc = os.path.join(tmpDir, "dirac-install.py")
            shutil.copyfile("dirac-install.py", install_loc)
            os.chmod(install_loc, 0o755)

            infoDict = None
            if os.path.isfile("pilot.json"):  # if this is a pilot 3 this file should be found
                with io.open("pilot.json") as pj:
                    infoDict = json.load(pj)

            # Extra Wrapper (Container DIRAC installer)
            wrapSubs = {
                "next_wrapper": wrapperPath,
                "install_args": self.__getInstallFlags(infoDict),
                "config_args": self.__getConfigFlags(infoDict),
            }
            CONTAINER_WRAPPER = CONTAINER_WRAPPER_INSTALL

        else:  # In case we don't (re)install DIRAC
            wrapSubs = {
                "next_wrapper": wrapperPath,
                "dirac_env_var": os.environ.get("DIRAC", ""),
                "diracos_env_var": os.environ.get("DIRACOS", ""),
            }
            if six.PY2:
                shutil.copyfile(
                    os.path.join(self.__findInstallBaseDir(), "bashrc"),
                    os.path.join(tmpDir, "bashrc"),
                )
                wrapSubs["rc_script"] = "bashrc"
            else:
                wrapSubs["rc_script"] = os.path.join(self.__findInstallBaseDir(), "diracosrc")
            shutil.copyfile("pilot.cfg", os.path.join(tmpDir, "pilot.cfg"))
            CONTAINER_WRAPPER = CONTAINER_WRAPPER_NO_INSTALL

        wrapLoc = os.path.join(tmpDir, "dirac_container.sh")
        rawfd = os.open(wrapLoc, os.O_WRONLY | os.O_CREAT, 0o700)
        fd = os.fdopen(rawfd, "w")
        fd.write(CONTAINER_WRAPPER % wrapSubs)
        fd.close()

        ret = S_OK()
        ret["baseDir"] = baseDir
        ret["tmpDir"] = tmpDir
        if proxy:
            ret["proxyLocation"] = proxyLoc
        return ret

    def __deleteWorkArea(self, baseDir):
        """Deletes the container work area (baseDir path) unless 'KeepWorkArea'
        option is set. Returns None.
        """
        if self.ceParameters.get("KeepWorkArea", False):
            return
        # We can't really do anything about errors: The tree should be fully owned
        # by the pilot user, so we don't expect any permissions problems.
        shutil.rmtree(baseDir, ignore_errors=True)

    def __getEnv(self):
        """Gets the environment for use within the container.
        We blank almost everything to prevent contamination from the host system.
        """
        payloadEnv = {}
        if "TERM" in os.environ:
            payloadEnv["TERM"] = os.environ["TERM"]
        payloadEnv["TMP"] = "/tmp"
        payloadEnv["TMPDIR"] = "/tmp"
        payloadEnv["X509_USER_PROXY"] = os.path.join(self.__innerdir, "proxy")
        return payloadEnv

    @staticmethod
    def __checkResult(tmpDir):
        """Gets the result of the payload command and returns it."""
        # The wrapper writes the inner job return code to "retcode"
        # in the working directory.
        try:
            with open(os.path.join(tmpDir, "retcode"), "rt") as fp:
                retCode = int(fp.read())
        except (IOError, ValueError):
            # Something failed while trying to get the return code
            result = S_ERROR("Failed to get return code from inner wrapper")
            result["ReschedulePayload"] = True
            return result

        result = S_OK()
        if retCode:
            # This is the one case where we don't reschedule:
            # An actual failure of the inner payload for some reason
            result = S_ERROR("Command failed with exit code %d" % retCode)
        return result

    def submitJob(self, executableFile, proxy=None, **kwargs):
        """Start a container for a job.
        executableFile is ignored. A new wrapper suitable for running in a
        container is created from jobDesc.
        """
        rootImage = self.__root

        # Check that singularity is available
        if not self.__hasSingularity():
            self.log.error("Singularity is not installed on PATH.")
            result = S_ERROR("Failed to find singularity ")
            result["ReschedulePayload"] = True
            return result

        self.log.info("Creating singularity container")

        # Start by making the directory for the container
        ret = self.__createWorkArea(kwargs.get("jobDesc"), kwargs.get("log"), kwargs.get("logLevel", "INFO"), proxy)
        if not ret["OK"]:
            return ret
        baseDir = ret["baseDir"]
        tmpDir = ret["tmpDir"]

        if proxy:
            payloadProxyLoc = ret["proxyLocation"]

            # Now we have to set-up payload proxy renewal for the container
            # This is fairly easy as it remains visible on the host filesystem
            result = gThreadScheduler.addPeriodicTask(
                self.proxyCheckPeriod, self._monitorProxy, taskArgs=(payloadProxyLoc,), executions=0, elapsedTime=0
            )
            if result["OK"]:
                renewTask = result["Value"]
            else:
                self.log.warn("Failed to start proxy renewal task")
                renewTask = None

        # Very simple accounting
        self.__submittedJobs += 1
        self.__runningJobs += 1

        # Now prepare start singularity
        # Mount /cvmfs in if it exists on the host
        withCVMFS = os.path.isdir("/cvmfs")
        innerCmd = os.path.join(self.__innerdir, "dirac_container.sh")
        cmd = [self.__singularityBin, "exec"]
        cmd.extend(["--contain"])  # use minimal /dev and empty other directories (e.g. /tmp and $HOME)
        cmd.extend(["--ipc", "--pid"])  # run container in new IPC and PID namespaces
        cmd.extend(["--workdir", baseDir])  # working directory to be used for /tmp, /var/tmp and $HOME
        if self.__hasUserNS():
            cmd.append("--userns")
        if withCVMFS:
            cmd.extend(["--bind", "/cvmfs"])
        if not self.__installDIRACInContainer:
            cmd.extend(["--bind", "{0}:{0}:ro".format(self.__findInstallBaseDir())])
        if "ContainerBind" in self.ceParameters:
            bindPaths = self.ceParameters["ContainerBind"].split(",")
            for bindPath in bindPaths:
                if len(bindPath.split(":::")) == 1:
                    cmd.extend(["--bind", bindPath.strip()])
                elif len(bindPath.split(":::")) in [2, 3]:
                    cmd.extend(["--bind", ":".join([bp.strip() for bp in bindPath.split(":::")])])
        if "ContainerOptions" in self.ceParameters:
            containerOpts = self.ceParameters["ContainerOptions"].split(",")
            for opt in containerOpts:
                cmd.extend([opt.strip()])
        if os.path.isdir(rootImage) or os.path.isfile(rootImage):
            cmd.extend([rootImage, innerCmd])
        else:
            # if we are here is because there's no image, or it is not accessible (e.g. not on CVMFS)
            self.log.error("Singularity image to exec not found: ", rootImage)
            result = S_ERROR("Failed to find singularity image to exec")
            result["ReschedulePayload"] = True
            return result

        self.log.debug("Execute singularity command: %s" % cmd)
        self.log.debug("Execute singularity env: %s" % self.__getEnv())
        result = systemCall(0, cmd, callbackFunction=self.sendOutput, env=self.__getEnv())

        self.__runningJobs -= 1

        if not result["OK"]:
            self.log.error("Fail to run Singularity", result["Message"])
            if proxy and renewTask:
                gThreadScheduler.removeTask(renewTask)
            self.__deleteWorkArea(baseDir)
            result = S_ERROR("Error running singularity command")
            result["ReschedulePayload"] = True
            return result

        result = self.__checkResult(tmpDir)
        if proxy and renewTask:
            gThreadScheduler.removeTask(renewTask)
        self.__deleteWorkArea(baseDir)
        return result

    def getCEStatus(self):
        """Method to return information on running and pending jobs."""
        result = S_OK()
        result["SubmittedJobs"] = self.__submittedJobs
        result["RunningJobs"] = self.__runningJobs
        result["WaitingJobs"] = 0
        # processors
        result["AvailableProcessors"] = self.processors
        return result



"""
Service handler for FTS3DB using DISET

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN FTS3Manager
  :end-before: ##END
  :dedent: 2
  :caption: FTS3Manager options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six

# from DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername
from DIRAC.Core.DISET.RequestHandler import RequestHandler, getServiceOption
from DIRAC.Core.Security.Properties import FULL_DELEGATION, LIMITED_DELEGATION
from DIRAC.Core.Utilities import DErrno
from DIRAC.Core.Utilities.JEncode import strToIntDict


from DIRAC.DataManagementSystem.DB.FTS3DB import FTS3DB
from DIRAC.Core.Utilities.JEncode import encode, decode

########################################################################


class FTS3ManagerHandlerMixin(object):
    """
    .. class:: FTS3ManagerHandlerMixin

    """

    fts3db = None

    @classmethod
    def initializeHandler(cls, serviceInfoDict):
        """initialize handler"""
        try:
            maxThreads = getServiceOption(serviceInfoDict, "MaxThreads", 15)
            cls.fts3db = FTS3DB(pool_size=maxThreads)
        except RuntimeError as error:
            gLogger.exception(error)
            return S_ERROR(error)

        # # create tables for empty db
        return cls.fts3db.createTables()

    @staticmethod
    def _isAllowed(opObj, remoteCredentials):
        """
        Make sure the client is allowed to persist an operation
        (FULL_DELEGATION or LIMITED_DELEGATION). This is the case of pilots,
        the RequestExecutingAgent or the FTS3Agent

        :param opObj: the FTS3Operation object
        :param remoteCredentials: credentials from the clients

        :returns: True if everything is fine, False otherwise
        """

        credDN = remoteCredentials["DN"]
        credGroup = remoteCredentials["group"]
        credProperties = remoteCredentials["properties"]

        # First, get the DN matching the username
        res = getDNForUsername(opObj.username)
        # if we have an error, do not allow
        if not res["OK"]:
            gLogger.error("Error retrieving DN for username", res)
            return False

        # List of DN matching the username
        dnList = res["Value"]

        # If the credentials in the Request match those from the credentials, it's OK
        if credDN in dnList and opObj.userGroup == credGroup:
            return True

        # From here, something/someone is putting a request on behalf of someone else

        # Only allow this if the credentials have Full or Limited delegation properties

        if FULL_DELEGATION in credProperties or LIMITED_DELEGATION in credProperties:
            return True

        return False

    types_persistOperation = [six.string_types]

    def export_persistOperation(self, opJSON):
        """update or insert request into db

        :param opJSON: json string representing the operation

        :return: OperationID
        """

        opObj, _size = decode(opJSON)

        isAuthorized = self._isAllowed(opObj, self.getRemoteCredentials())

        if not isAuthorized:
            return S_ERROR(DErrno.ENOAUTH, "Credentials in the requests are not allowed")

        return self.fts3db.persistOperation(opObj)

    types_getOperation = [six.integer_types]

    @classmethod
    def export_getOperation(cls, operationID):
        """Get the FTS3Operation from the database

        :param operationID: ID of the operation

        :return: the FTS3Operation JSON string matching
        """
        getOperation = cls.fts3db.getOperation(operationID)
        if not getOperation["OK"]:
            gLogger.error("getOperation: %s" % getOperation["Message"])
            return getOperation

        getOperation = getOperation["Value"]
        opJSON = encode(getOperation)
        return S_OK(opJSON)

    types_getActiveJobs = [six.integer_types, [type(None)] + [six.string_types], six.string_types]

    @classmethod
    def export_getActiveJobs(cls, limit, lastMonitor, jobAssignmentTag):
        """Get all the FTSJobs that are not in a final state

        :param limit: max number of jobs to retrieve
        :param jobAssignmentTag: tag to put in the DB
        :param lastMonitor: jobs monitored earlier than the given date
        :return: json list of FTS3Job
        """

        res = cls.fts3db.getActiveJobs(limit=limit, lastMonitor=lastMonitor, jobAssignmentTag=jobAssignmentTag)
        if not res["OK"]:
            return res

        activeJobs = res["Value"]
        activeJobsJSON = encode(activeJobs)

        return S_OK(activeJobsJSON)

    types_updateFileStatus = [dict]

    @classmethod
    def export_updateFileStatus(cls, fileStatusDict, ftsGUID):
        """Update the file ftsStatus and error

        :param fileStatusDict: { fileID : { status , error } }
        :param ftsGUID: (not mandatory) If specified, only update the rows where the ftsGUID matches this value.
        """
        fileStatusDict = strToIntDict(fileStatusDict)
        return cls.fts3db.updateFileStatus(fileStatusDict, ftsGUID)

    types_updateJobStatus = [dict]

    @classmethod
    def export_updateJobStatus(cls, jobStatusDict):
        """Update the job Status and error

        :param jobStatusDict: { jobID : { status , error } }
        """
        jobStatusDict = strToIntDict(jobStatusDict)
        return cls.fts3db.updateJobStatus(jobStatusDict)

    types_getNonFinishedOperations = [six.integer_types, six.string_types]

    @classmethod
    def export_getNonFinishedOperations(cls, limit, operationAssignmentTag):
        """Get all the FTS3Operations that are missing a callback, i.e.
        in 'Processed' state

        :param limit: max number of operations to retrieve
        :return: json list of FTS3Operation
        """

        res = cls.fts3db.getNonFinishedOperations(limit=limit, operationAssignmentTag=operationAssignmentTag)
        if not res["OK"]:
            return res

        nonFinishedOperations = res["Value"]
        nonFinishedOperationsJSON = encode(nonFinishedOperations)

        return S_OK(nonFinishedOperationsJSON)

    types_getOperationsFromRMSOpID = [six.integer_types]

    @classmethod
    def export_getOperationsFromRMSOpID(cls, rmsOpID):
        """Get the FTS3Operation associated to a given rmsOpID

        :param rmsOpID: ID of the operation in the RMS

        :return: JSON encoded list of FTS3Operations
        """
        res = cls.fts3db.getOperationsFromRMSOpID(rmsOpID)
        if not res["OK"]:
            gLogger.error("getOperationsFromRMSOpID: %s" % res["Message"])
            return res

        operations = res["Value"]
        opsJSON = encode(operations)

        return S_OK(opsJSON)


class FTS3ManagerHandler(FTS3ManagerHandlerMixin, RequestHandler):
    """DISET handler for FTS3Manager"""

    pass



#!/usr/bin/env python
"""
Create a DIRAC RemoveReplica|RemoveFile request to be executed by the RMS
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "ea64b42 (2012-07-29 16:45:05 +0200) ricardo <Ricardo.Graciani@gmail.com>"

import os
from hashlib import md5
import time
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC.Core.Utilities.List import breakListIntoChunks


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(" SE:   StorageElement|All")
    Script.registerArgument(["LFN:  LFN or file containing a List of LFNs"])
    Script.parseCommandLine(ignoreErrors=False)

    # parseCommandLine show help when mandatory arguments are not specified or incorrect argument
    args = Script.getPositionalArgs()

    targetSE = args.pop(0)

    lfns = []
    for inputFileName in args:
        if os.path.exists(inputFileName):
            with open(inputFileName, "r") as inputFile:
                string = inputFile.read()
            lfns.extend([lfn.strip() for lfn in string.splitlines()])
        else:
            lfns.append(inputFileName)

    from DIRAC.Resources.Storage.StorageElement import StorageElement
    import DIRAC

    # Check is provided SE is OK
    if targetSE != "All":
        se = StorageElement(targetSE)
        if not se.valid:
            print(se.errorReason)
            print()
            Script.showHelp()

    from DIRAC.RequestManagementSystem.Client.Request import Request
    from DIRAC.RequestManagementSystem.Client.Operation import Operation
    from DIRAC.RequestManagementSystem.Client.File import File
    from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
    from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

    reqClient = ReqClient()
    fc = FileCatalog()

    requestOperation = "RemoveReplica"
    if targetSE == "All":
        requestOperation = "RemoveFile"

    for lfnList in breakListIntoChunks(lfns, 100):

        oRequest = Request()
        requestName = "%s_%s" % (
            md5(repr(time.time()).encode()).hexdigest()[:16],
            md5(repr(time.time()).encode()).hexdigest()[:16],
        )
        oRequest.RequestName = requestName

        oOperation = Operation()
        oOperation.Type = requestOperation
        oOperation.TargetSE = targetSE

        res = fc.getFileMetadata(lfnList)
        if not res["OK"]:
            print("Can't get file metadata: %s" % res["Message"])
            DIRAC.exit(1)
        if res["Value"]["Failed"]:
            print("Could not get the file metadata of the following, so skipping them:")
            for fFile in res["Value"]["Failed"]:
                print(fFile)

        lfnMetadata = res["Value"]["Successful"]

        for lfn in lfnMetadata:
            rarFile = File()
            rarFile.LFN = lfn
            rarFile.Size = lfnMetadata[lfn]["Size"]
            rarFile.Checksum = lfnMetadata[lfn]["Checksum"]
            rarFile.GUID = lfnMetadata[lfn]["GUID"]
            rarFile.ChecksumType = "ADLER32"
            oOperation.addFile(rarFile)

        oRequest.addOperation(oOperation)

        isValid = RequestValidator().validate(oRequest)
        if not isValid["OK"]:
            print("Request is not valid: ", isValid["Message"])
            DIRAC.exit(1)

        result = reqClient.putRequest(oRequest)
        if result["OK"]:
            print("Request %d Submitted" % result["Value"])
        else:
            print("Failed to submit Request: ", result["Message"])


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from DIRAC.FrameworkSystem.private.monitoring.ColorGenerator import ColorGenerator


def test_getHexColor():
    cg = ColorGenerator()
    for i in range(2):
        assert cg.getHexColor() == "0000FF"
        assert cg.getHexColor() == "00FF00"
        assert cg.getHexColor() == "FF0000"
        assert cg.getHexColor() == "00FFFF"
        assert cg.getHexColor() == "FF00FF"
        assert cg.getHexColor() == "FFFF00"
        assert cg.getHexColor() == "007F7F"
        assert cg.getHexColor() == "7F007F"
        assert cg.getHexColor() == "7F7F00"
        assert cg.getHexColor() == "007FFF"
        assert cg.getHexColor() == "00FF7F"
        assert cg.getHexColor() == "7F7F7F"
        assert cg.getHexColor() == "7F00FF"
        assert cg.getHexColor() == "FF007F"
        assert cg.getHexColor() == "7F7FFF"
        for j in range(1000):
            assert len(cg.getHexColor()) == 6
        cg.reset()


def test_getFloatColor():
    cg = ColorGenerator()
    cg.reset()
    for i in range(2):
        assert cg.getFloatColor() == (0.0, 0.0, 1.0)
        assert cg.getFloatColor() == (0.0, 1.0, 0.0)
        assert cg.getFloatColor() == (1.0, 0.0, 0.0)
        assert cg.getFloatColor() == (0.0, 1.0, 1.0)
        assert cg.getFloatColor() == (1.0, 0.0, 1.0)
        assert cg.getFloatColor() == (1.0, 1.0, 0.0)
        assert cg.getFloatColor() == (0.0, 0.5, 0.5)
        assert cg.getFloatColor() == (0.5, 0.0, 0.5)
        assert cg.getFloatColor() == (0.5, 0.5, 0.0)
        assert cg.getFloatColor() == (0.0, 0.5, 1.0)
        assert cg.getFloatColor() == (0.0, 1.0, 0.5)
        assert cg.getFloatColor() == (0.5, 0.5, 0.5)
        assert cg.getFloatColor() == (0.5, 0.0, 1.0)
        assert cg.getFloatColor() == (1.0, 0.0, 0.5)
        assert cg.getFloatColor() == (0.5, 0.5, 1.0)
        for j in range(1000):
            result = cg.getFloatColor()
            assert len(result) == 3
            assert all(isinstance(x, float) for x in result)
        cg.reset()



########################################################################
# File :    ModuleFactory.py
# Author :  Stuart Paterson
########################################################################
"""  The Module Factory instantiates a given Module based on a given input
     string and set of arguments to be passed.  This allows for VO specific
     module utilities to be used in various contexts.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR, gLogger


class ModuleFactory(object):

    #############################################################################
    def __init__(self):
        """Standard constructor"""
        self.log = gLogger

    #############################################################################
    def getModule(self, importString, argumentsDict):
        """This method returns the Module instance given the import string and
        arguments dictionary.
        """
        try:
            moduleName = importString.split(".")[-1]
            modulePath = importString.replace(".%s" % (moduleName), "")
            importModule = __import__("%s.%s" % (modulePath, moduleName), globals(), locals(), [moduleName])
        except Exception as x:
            msg = "ModuleFactory could not import %s.%s" % (modulePath, moduleName)
            self.log.warn(x)
            self.log.warn(msg)
            return S_ERROR(msg)

        try:
            # FIXME: should we use imp module?
            moduleStr = "importModule.%s(argumentsDict)" % (moduleName)
            moduleInstance = eval(moduleStr)
        except Exception as x:
            msg = "ModuleFactory could not instantiate %s()" % (moduleName)
            self.log.warn(x)
            self.log.warn(msg)
            return S_ERROR(msg)

        return S_OK(moduleInstance)


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



""" Client to interact with the S3Gateway  """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Base.Client import Client, createClient


@createClient("DataManagement/S3Gateway")
class S3GatewayClient(Client):
    """Client code to the S3Gateway"""

    def __init__(self, url=None, **kwargs):
        """Constructor function."""
        super(S3GatewayClient, self).__init__(**kwargs)
        self.setServer("DataManagement/S3Gateway")
        if url:
            self.setServer(url)

    def createPresignedUrl(self, storageName, s3_method, urls, expiration=3600, **kwargs):
        """Generate a presigned URL for a given object, given method, and given storage
        Permissions are checked against the DFC

        :param storageName: SE name
        :param s3_method: name of the S3 client method we want to perform.
        :param urls: Iterable of urls. If s3_method is put_object, it must be a dict <url:fields> where fields
                     is a dictionary (see ~DIRAC.Resources.Storage.S3Storage.S3Storage.createPresignedUrl)
        :param expiration: duration of the token
        """

        return self._getRPC(**kwargs).createPresignedUrl(storageName, s3_method, urls, expiration)
